﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.server;

namespace simpledb.buffer
{
    /// <summary>
    /// An individual buffer.
    /// A buffer wraps a page and stores information about its status,
    /// such as the disk block associated with the page,
    /// the number of times the block has been pinned,
    /// whether the contents of the page have been modified,
    /// and if so, the id of the modifying transaction and
    /// the LSN of the corresponding log record.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class Buffer
    {
        private Page contents = new Page();
        private Block blk = null;
        private int pins = 0;
        private int modifiedBy = -1;  // negative means not modified
        private int logSequenceNumber = -1; // negative means no corresponding log record

        /// <summary>
        /// Returns the integer value at the specified offset of the
        /// buffer's page.
        /// If an integer was not stored at that location,
        /// the behavior of the method is unpredictable.
        /// </summary>
        /// <param name="offset">the byte offset of the page</param>  
        /// <returns>the integer value at that offset</returns>
        public int getInt(int offset)
        {
            return contents.getInt(offset);
        }

        /// <summary>
        /// Returns the string value at the specified offset of the
        /// buffer's page.
        /// If a string was not stored at that location,
        /// the behavior of the method is unpredictable.
        /// </summary>
        /// <param name="offset">the byte offset of the page</param>  
        /// <returns>the string value at that offset</returns>
        public string getString(int offset)
        {
            return contents.getString(offset);
        }

        /// <summary>
        /// Writes an integer to the specified offset of the
        /// buffer's page.
        /// This method assumes that the transaction has already
        /// written an appropriate log record.
        /// The buffer saves the id of the transaction
        /// and the LSN of the log record.
        /// A negative lsn value indicates that a log record
        /// was not necessary.
        /// </summary>
        /// <param name="offset">the byte offset within the page</param>  
        /// <param name="val">the new integer value to be written</param>  
        /// <param name="txnum">the id of the transaction performing the modification</param>  
        /// <param name="lsn">the LSN of the corresponding log record</param>
        public void setInt(int offset, int val, int txnum, int lsn)
        {
            modifiedBy = txnum;
            if (lsn >= 0)
                logSequenceNumber = lsn;
            contents.setInt(offset, val);
        }

        /// <summary>
        /// Writes a string to the specified offset of the
        /// buffer's page.
        /// This method assumes that the transaction has already
        /// written an appropriate log record.
        /// A negative lsn value indicates that a log record
        /// was not necessary.
        /// The buffer saves the id of the transaction
        /// and the LSN of the log record.
        /// </summary>
        /// <param name="offset">the byte offset within the page</param>  
        /// <param name="val">the new string value to be written</param>  
        /// <param name="txnum">the id of the transaction performing the modification</param>  
        /// <param name="lsn">the LSN of the corresponding log record</param>     
        public void setString(int offset, string val, int txnum, int lsn)
        {
            modifiedBy = txnum;
            if (lsn >= 0)
                logSequenceNumber = lsn;
            contents.setString(offset, val);
        }

        /// <summary>
        /// Returns a reference to the disk block
        /// that the buffer is pinned to.
        /// </summary>
        /// <returns>a reference to a disk block</returns>
        public Block block()
        {
            return blk;
        }

        /// <summary>
        /// Writes the page to its disk block if the
        /// page is dirty.
        /// The method ensures that the corresponding log
        /// record has been written to disk prior to writing
        /// the page to disk. 
        /// </summary>
        internal void flush()
        {
            if (modifiedBy >= 0)
            {
                SimpleDB.logMgr().flush(logSequenceNumber);
                contents.write(blk);
                modifiedBy = -1;
            }
        }

        /// <summary>
        /// Increases the buffer's pin count. 
        /// </summary>
        internal void pin()
        {
            pins++;
        }

        /// <summary>
        /// Increases the buffer's pin count. 
        /// </summary>
        internal void unpin()
        {
            pins--;
        }

        /// <summary>
        /// Returns true if the buffer is currently pinned
        /// (that is, if it has a nonzero pin count).         
        /// </summary>
        /// <returns>true if the buffer is pinned</returns> 
        internal bool isPinned()
        {
            return pins > 0;
        }

        /// <summary>
        /// Returns true if the buffer is dirty
        /// due to a modification by the specified transaction.
        /// </summary> 
        /// <param name="txnum ">the id of the transaction</param> 
        /// <returns> true if the transaction modified the buffer</returns>
        internal bool isModifiedBy(int txnum)
        {
            return modifiedBy == txnum;
        }

        /// <summary>
        /// Reads the contents of the specified block into
        /// the buffer's page.
        /// If the buffer was dirty, then the contents
        /// of the previous page are first written to disk.
        /// </summary>
        /// <param name="b">a reference to the data block</param>  
        internal void assignToBlock(Block b)
        {
            flush();
            blk = b;
            contents.read(blk);
            pins = 0;
        }

        /// <summary>
        /// Initializes the buffer's page according to the specified formatter,
        /// and appends the page to the specified file.
        /// If the buffer was dirty, then the contents
        /// of the previous page are first written to disk.
        /// </summary>
        /// <param name="filename"> the name of the file</param> 
        /// <param name="fmtr">a page formatter, used to initialize the page</param> 
        internal void assignToNew(string filename, PageFormatter fmtr)
        {
            flush();
            fmtr.format(contents);
            blk = contents.append(filename);
            pins = 0;
        }
    }
}
