﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System;
using System.Collections.Generic;
using System.IO;

namespace simpledb.file
{
    /// <summary>
    /// The SimpleDB file manager.
    /// The database system stores its data as files within a specified directory.
    /// The file manager provides methods for reading the contents of
    /// a file block to a Java byte buffer,
    /// writing the contents of a byte buffer to a file block,
    /// and appending the contents of a byte buffer to the end of a file.
    /// These methods are called exclusively by the class <see cref="Page"/>,
    /// and are thus package-private.
    /// The class also contains two public methods:
    /// Method <see cref="isNew"/> is called during system initialization by <see cref="simpledb.server.SimpleDB.init"/> .
    /// Method <see cref="size"/> is called by the log manager and transaction manager to
    /// determine the end of the file.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class FileMgr
    {
        private string dbDirectory;
        private bool isnew;
        private Dictionary<string, FileStream> openFiles = new Dictionary<string, FileStream>();
        private object threadLock = new object();

        /// <summary>
        /// Creates a file manager for the specified database.
        /// The database will be stored in a folder of that name
        /// in the user's home directory.
        /// If the folder does not exist, then a folder containing
        /// an empty database is created automatically.
        /// Files for all temporary tables (i.e. tables beginning with "temp") are deleted.
        /// </summary>
        /// <param name="dbname">the name of the directory that holds the database</param>
        public FileMgr(string dbname)
        {
            string homedir = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            dbDirectory = Path.Combine(homedir, dbname);
            DirectoryInfo di = new DirectoryInfo(dbDirectory);
            isnew = !di.Exists;

            // create the directory if the database is new
            if (isnew)
                di.Create();

            // remove any leftover temporary tables
            foreach (FileInfo fi in di.GetFiles())
                if (fi.Name.StartsWith("temp"))
                    fi.Delete();
        }

        /// <summary>
        /// Reads the contents of a disk block into a bytebuffer.
        /// </summary>
        /// <param name="blk">a reference to a disk block</param>
        /// <param name="bb">the byte[] buffer</param>
        internal void read(Block blk, byte[] bb)
        {
            lock (threadLock)
            {
                try
                {
                    bb.Initialize();
                    FileStream fs = getFile(blk.fileName());
                    fs.Seek(blk.number() * Page.BLOCK_SIZE, SeekOrigin.Begin);
                    fs.Read(bb, 0, Page.BLOCK_SIZE);
                }
                catch (IOException)
                {
                    throw new Exception("cannot read block " + blk);
                }
            }
        }

        /// <summary>
        /// Writes the contents of a bytebuffer into a disk block. 
        /// </summary>
        /// <param name="blk">a reference to a disk block</param>
        /// <param name="bb">the byte[] buffer</param>
        internal void write(Block blk, byte[] bb)
        {
            lock (threadLock)
            {
                try
                {
                    FileStream fs = getFile(blk.fileName());
                    fs.Seek(blk.number() * Page.BLOCK_SIZE, SeekOrigin.Begin);
                    fs.Write(bb, 0, Page.BLOCK_SIZE);
                }
                catch (IOException)
                {
                    throw new Exception("cannot write block " + blk);
                }
            }
        }

        /// <summary>
        /// Appends the contents of a bytebuffer to the end 
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <param name="bb">the byte[] buffer</param>
        /// <returns>a reference to the newly-created block.</returns>
        internal Block append(string filename, byte[] bb)
        {
            lock (threadLock)
            {
                int newblknum = size(filename);
                Block blk = new Block(filename, newblknum);
                write(blk, bb);
                return blk;
            }
        }

        /// <summary>
        /// Returns the number of blocks in the specified file.
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <returns>the number of blocks in the file</returns>
        public int size(string filename)
        {
            lock (threadLock)
            {
                try
                {
                    FileStream fs = getFile(filename);
                    return (int)(fs.Length / Page.BLOCK_SIZE);
                }
                catch (IOException)
                {
                    throw new Exception("cannot access " + filename);
                }
            }
        }

        /// <summary>
        /// Returns a boolean indicating whether the file manager
        /// had to create a new database directory. 
        /// </summary>
        /// <returns>true if the database is new</returns>
        public bool isNew()
        {
            return isnew;
        }

        /// <summary>
        /// Returns the file stream for the specified filename.
        /// The file stream is stored in a map keyed on the filename.
        /// If the file is not open, then it is opened and the file channel
        /// is added to the map.
        /// </summary>
        /// <param name="filename">the specified filename</param>
        /// <returns>the file stream associated with the open file.</returns> 
        /// <exception cref="IOException" />
        private FileStream getFile(string filename)
        {
            FileStream fs;

            if (openFiles.ContainsKey(filename))
            {
                fs = openFiles[filename];
            }
            else
            {
                string dbTable = Path.Combine(dbDirectory, filename);
                fs = new FileStream(dbTable, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.None, 8, FileOptions.WriteThrough);
                openFiles.Add(filename, fs);
            }
            return fs;
        }
    }
}