﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.tx;
using simpledb.record;
using simpledb.query;

namespace simpledb.index.btree
{
    /// <summary>
    /// A B-tree implementation of the Index interface
    /// </summary>.
    /// <author>Edward Sciore</author> 
    public class BTreeIndex : Index
    {
        private Transaction tx;
        private TableInfo dirTi, leafTi;
        private BTreeLeaf leaf = null;
        private Block rootblk;

        /// <summary>
        /// Opens a B-tree index for the specified index.
        /// The method determines the appropriate files
        /// for the leaf and directory records,
        /// creating them if they did not exist.
        /// </summary>
        /// <param name="idxname">the name of the index</param>
        /// <param name="leafsch">the schema of the leaf index records</param>
        /// <param name="tx">the calling transaction</param>
        public BTreeIndex(string idxname, Schema leafsch, Transaction tx)
        {
            this.tx = tx;
            // deal with the leaves
            string leaftbl = idxname + "leaf";
            leafTi = new TableInfo(leaftbl, leafsch);
            if (tx.size(leafTi.fileName()) == 0)
                tx.append(leafTi.fileName(), new BTPageFormatter(leafTi, -1));

            // deal with the directory
            Schema dirsch = new Schema();
            dirsch.add("block", leafsch);
            dirsch.add("dataval", leafsch);
            string dirtbl = idxname + "dir";
            dirTi = new TableInfo(dirtbl, dirsch);
            rootblk = new Block(dirTi.fileName(), 0);
            if (tx.size(dirTi.fileName()) == 0)
                // create new root block
                tx.append(dirTi.fileName(), new BTPageFormatter(dirTi, 0));
            BTreePage page = new BTreePage(rootblk, dirTi, tx);
            if (page.getNumRecs() == 0)
            {
                // insert initial directory entry
                int fldtype = dirsch.type("dataval");
                Constant minval = (fldtype == Schema.INTEGER) ?
                    (Constant)new IntConstant(int.MaxValue) : (Constant)new StringConstant("");
                page.insertDir(0, minval, 0);
            }
            page.close();
        }

        /// <summary>
        /// Traverses the directory to find the leaf block corresponding
        /// to the specified search key.
        /// The method then opens a page for that leaf block, and
        /// positions the page before the first record (if any)
        /// having that search key.
        /// The leaf page is kept open, for use by the methods next
        /// and getDataRid.
        /// </summary>
        /// <seealso cref="simpledb.index.Index.beforeFirst"/> 
        public void beforeFirst(Constant searchkey)
        {
            close();
            BTreeDir root = new BTreeDir(rootblk, dirTi, tx);
            int blknum = root.search(searchkey);
            root.close();
            Block leafblk = new Block(leafTi.fileName(), blknum);
            leaf = new BTreeLeaf(leafblk, leafTi, searchkey, tx);
        }

        /// <summary>
        /// Moves to the next leaf record having the
        /// previously-specified search key.
        /// Returns false if there are no more such leaf records
        /// </summary>
        /// <seealso cref="simpledb.index.Index.next"/> 
        public bool next()
        {
            return leaf.next();
        }

        /// <summary>
        /// Returns the dataRID value from the current leaf record
        /// </summary>
        /// <seealso cref="simpledb.index.Index.getDataRid"/> 
        public RID getDataRid()
        {
            return leaf.getDataRid();
        }

        /// <summary>
        /// Inserts the specified record into the index.
        /// The method first traverses the directory to find
        /// the appropriate leaf page; then it inserts
        /// the record into the leaf.
        /// If the insertion causes the leaf to split, then
        /// the method calls insert on the root,
        /// passing it the directory entry of the new leaf page.
        /// If the root node splits, then makeNewRoot is called.
        /// </summary>
        /// <seealso cref="simpledb.index.Index.insert"/> 
        public void insert(Constant dataval, RID datarid)
        {
            beforeFirst(dataval);
            DirEntry e = leaf.insert(datarid);
            leaf.close();
            if (e == null)
                return;
            BTreeDir root = new BTreeDir(rootblk, dirTi, tx);
            DirEntry e2 = root.insert(e);
            if (e2 != null)
                root.makeNewRoot(e2);
            root.close();
        }

        /// <summary>
        /// Deletes the specified index record.
        /// The method first traverses the directory to find
        /// the leaf page containing that record; then it
        /// deletes the record from the page
        /// </summary>
        /// <seealso cref="simpledb.index.Index.delete"/> 
        public void delete(Constant dataval, RID datarid)
        {
            beforeFirst(dataval);
            leaf.delete(datarid);
            leaf.close();
        }

        /// <summary>
        /// Closes the index by closing its open leaf page,
        /// if necessary.
        /// </summary>
        /// <seealso cref="simpledb.index.Index.close"/> 
        public void close()
        {
            if (leaf != null)
                leaf.close();
        }

        /// <summary>
        /// Estimates the number of block accesses
        /// required to find all index records having
        /// a particular search key  
        /// </summary>
        /// <param name="numblocks">the number of blocks in the B-tree directory</param>
        /// <param name="rpb">the number of index entries per block</param>
        /// <returns>the estimated traversal cost</returns>
        public static int searchCost(int numblocks, int rpb)
        {
            return 1 + (int)(System.Math.Log(numblocks) / System.Math.Log(rpb));
        }
    }
}
