﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.tx;
using simpledb.record;
using simpledb.query;

namespace simpledb.index.btree
{
    /// <summary>
    /// An object that holds the contents of a B-tree leaf block.
    /// </summary>
    /// <author>Edward Sciore</author> 
    public class BTreeLeaf
    {
        private TableInfo ti;
        private Transaction tx;
        private Constant searchkey;
        private BTreePage contents;
        private int currentslot;

        /// <summary>
        /// Opens a page to hold the specified leaf block.
        /// The page is positioned immediately before the first record
        /// having the specified search key (if any).
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        /// <param name="ti">the metadata of the B-tree leaf file</param>
        /// <param name="searchkey">the search key value</param>
        /// <param name="tx">the calling transaction</param>
        public BTreeLeaf(Block blk, TableInfo ti, Constant searchkey, Transaction tx)
        {
            this.ti = ti;
            this.tx = tx;
            this.searchkey = searchkey;
            contents = new BTreePage(blk, ti, tx);
            currentslot = contents.findSlotBefore(searchkey);
        }

        /// <summary>
        /// Closes the leaf page 
        /// </summary>
        public void close()
        {
            contents.close();
        }

        /// <summary>
        /// Moves to the next leaf record having the 
        /// previously-specified search key.
        /// Returns false if there is no more such records.
        /// </summary>
        /// <returns>false if there are no more leaf records for the search key</returns>
        public bool next()
        {
            currentslot++;
            if (currentslot >= contents.getNumRecs())
                return tryOverflow();
            else if (contents.getDataVal(currentslot).Equals(searchkey))
                return true;
            else
                return tryOverflow();
        }

        /// <summary>
        /// Returns the dataRID value of the current leaf record.
        /// </summary>
        /// <returns>the dataRID of the current record</returns>
        public RID getDataRid()
        {
            return contents.getDataRid(currentslot);
        }

        /// <summary>
        /// Deletes the leaf record having the specified dataRID
        /// </summary>
        /// <param name="datarid">the dataRId whose record is to be deleted</param>
        public void delete(RID datarid)
        {
            while (next())
                if (getDataRid().Equals(datarid))
                {
                    contents.delete(currentslot);
                    return;
                }
        }

        /// <summary>
        /// Inserts a new leaf record having the specified dataRID
        /// and the previously-specified search key.
        /// If the record does not fit in the page, then 
        /// the page splits and the method returns the
        /// directory entry for the new page;
        /// otherwise, the method returns null.  
        /// If all of the records in the page have the same dataval,
        /// then the block does not split; instead, all but one of the
        /// records are placed into an overflow block.
        /// </summary>
        /// <param name="datarid">the dataRID value of the new record</param>
        /// <returns>the directory entry of the newly-split page, if one exists</returns>
        public DirEntry insert(RID datarid)
        {
            currentslot++;
            contents.insertLeaf(currentslot, searchkey, datarid);
            if (!contents.isFull())
                return null;
            // else page is full, so split it
            Constant firstkey = contents.getDataVal(0);
            Constant lastkey = contents.getDataVal(contents.getNumRecs() - 1);
            if (lastkey.Equals(firstkey))
            {
                // create an overflow block to hold all but the first record
                Block newblk = contents.split(1, contents.getFlag());
                contents.setFlag(newblk.number());
                return null;
            }
            else
            {
                int splitpos = contents.getNumRecs() / 2;
                Constant splitkey = contents.getDataVal(splitpos);
                if (splitkey.Equals(firstkey))
                {
                    // move right, looking for the next key
                    while (contents.getDataVal(splitpos).Equals(splitkey))
                        splitpos++;
                    splitkey = contents.getDataVal(splitpos);
                }
                else
                {
                    // move left, looking for first entry having that key
                    while (contents.getDataVal(splitpos - 1).Equals(splitkey))
                        splitpos--;
                }
                Block newblk = contents.split(splitpos, -1);
                return new DirEntry(splitkey, newblk.number());
            }
        }
        private bool tryOverflow()
        {
            Constant firstkey = contents.getDataVal(0);
            int flag = contents.getFlag();
            if (!searchkey.Equals(firstkey) || flag < 0)
                return false;
            contents.close();
            Block nextblk = new Block(ti.fileName(), flag);
            contents = new BTreePage(nextblk, ti, tx);
            currentslot = 0;
            return true;
        }
    }
}
