﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.record;
using simpledb.query;
using simpledb.tx;

namespace simpledb.index.btree
{
    /// <summary>
    /// B-tree directory and leaf pages have many commonalities:
    /// in particular, their records are stored in sorted order, 
    /// and pages split when full.
    /// A BTreePage object contains this common functionality.
    /// <author>Edward Sciore</author> 
    /// </summary>
    public class BTreePage
    {
        private Block currentblk;
        private TableInfo ti;
        private Transaction tx;
        private int slotsize;

        /// <summary>
        /// Opens a page for the specified B-tree block.
        /// </summary>
        /// <param name="currentblk">a reference to the B-tree block</param>
        /// <param name="ti">the metadata for the particular B-tree file</param>
        /// <param name="tx">the calling transaction</param>
        public BTreePage(Block currentblk, TableInfo ti, Transaction tx)
        {
            this.currentblk = currentblk;
            this.ti = ti;
            this.tx = tx;
            slotsize = ti.recordLength();
            tx.pin(currentblk);
        }

        /// <summary>
        /// Calculates the position where the first record having
        /// the specified search key should be, then returns
        /// the position before it.
        /// </summary>
        /// <param name="searchkey">the search key</param>
        /// <returns>the position before where the search key goes</returns>
        public int findSlotBefore(Constant searchkey)
        {
            int slot = 0;
            while (slot < getNumRecs() && getDataVal(slot).CompareTo(searchkey) < 0)
                slot++;
            return slot - 1;

        }

        /// <summary>
        /// Closes the page by unpinning its buffer 
        /// </summary>
        public void close()
        {
            if (currentblk != null)
                tx.unpin(currentblk);
            currentblk = null;
        }

        /// <summary>
        /// Returns true if the block is full
        /// </summary>
        /// <returns>true if the block is full</returns>
        public bool isFull()
        {
            return slotpos(getNumRecs() + 1) >= Page.BLOCK_SIZE;
        }

        /// <summary>
        /// Splits the page at the specified position.
        /// A new page is created, and the records of the page
        /// starting at the split position are transferred to the new page
        /// </summary>
        /// <param name="splitpos">the split position</param>
        /// <param name="flag">the initial value of the flag field</param>
        /// <returns>the reference to the new block</returns>
        public Block split(int splitpos, int flag)
        {
            Block newblk = appendNew(flag);
            BTreePage newpage = new BTreePage(newblk, ti, tx);
            transferRecs(splitpos, newpage);
            newpage.setFlag(flag);
            newpage.close();
            return newblk;
        }

        /// <summary>
        /// Returns the dataval of the record at the specified slot.
        /// </summary>
        /// <param name="slot">the integer slot of an index record</param>
        /// <returns>the dataval of the record at that slot</returns>
        public Constant getDataVal(int slot)
        {
            return getVal(slot, "dataval");
        }

        /// <summary>
        /// Returns the value of the page's flag field
        /// </summary>
        /// <returns>the value of the page's flag field</returns>
        public int getFlag()
        {
            return tx.getInt(currentblk, 0);
        }

        /// <summary>
        /// Sets the page's flag field to the specified value
        /// </summary>
        /// <param name="val">the new value of the page flag</param>
        public void setFlag(int val)
        {
            tx.setInt(currentblk, 0, val);
        }

        /// <summary>
        /// Appends a new block to the end of the specified B-tree file,
        /// having the specified flag value.
        /// </summary>
        /// <param name="flag">the initial value of the flag</param>
        /// <returns>a reference to the newly-created block</returns>
        public Block appendNew(int flag)
        {
            return tx.append(ti.fileName(), new BTPageFormatter(ti, flag));
        }

        // Methods called only by BTreeDir

        /// <summary>
        /// Returns the block number stored in the index record 
        /// at the specified slot.
        /// </summary>
        /// <param name="slot">the slot of an index record</param>
        /// <returns>the block number stored in that record</returns>
        public int getChildNum(int slot)
        {
            return getInt(slot, "block");
        }

        /// <summary>
        /// Inserts a directory entry at the specified slot.
        /// </summary>
        /// <param name="slot">the slot of an index record</param>
        /// <param name="val">the dataval to be stored</param>
        /// <param name="blknum">the block number to be stored</param>
        public void insertDir(int slot, Constant val, int blknum)
        {
            insert(slot);
            setVal(slot, "dataval", val);
            setInt(slot, "block", blknum);
        }

        // Methods called only by BTreeLeaf

        /// <summary>
        /// Returns the dataRID value stored in the specified leaf index record.
        /// </summary>
        /// <param name="slot">the slot of the desired index record</param>
        /// <returns>the dataRID value store at that slot</returns>
        public RID getDataRid(int slot)
        {
            return new RID(getInt(slot, "block"), getInt(slot, "id"));
        }

        /// <summary>
        /// Inserts a leaf index record at the specified slot
        /// </summary>
        /// <param name="slot">the slot of the desired index record</param>
        /// <param name="val">the new dataval</param>
        /// <param name="rid">the new dataRID</param>
        public void insertLeaf(int slot, Constant val, RID rid)
        {
            insert(slot);
            setVal(slot, "dataval", val);
            setInt(slot, "block", rid.blockNumber());
            setInt(slot, "id", rid.Id());
        }

        /// <summary>
        /// Deletes the index record at the specified slot.
        /// </summary>
        /// <param name="slot">the slot of the deleted index record</param>
        public void delete(int slot)
        {
            for (int i = slot + 1; i < getNumRecs(); i++)
                copyRecord(i, i - 1);
            setNumRecs(getNumRecs() - 1);
            return;
        }

        /// <summary>
        /// Returns the number of index records in this page.
        /// </summary>
        /// <returns>the number of index records in this page</returns>
        public int getNumRecs()
        {
            return tx.getInt(currentblk, Page.INT_SIZE);
        }

        // Private methods

        private int getInt(int slot, string fldname)
        {
            int pos = fldpos(slot, fldname);
            return tx.getInt(currentblk, pos);
        }

        private string getString(int slot, string fldname)
        {
            int pos = fldpos(slot, fldname);
            return tx.getString(currentblk, pos);
        }

        private Constant getVal(int slot, string fldname)
        {
            int type = ti.schema().type(fldname);
            if (type == Schema.INTEGER)
                return new IntConstant(getInt(slot, fldname));
            else
                return new StringConstant(getString(slot, fldname));
        }

        private void setInt(int slot, string fldname, int val)
        {
            int pos = fldpos(slot, fldname);
            tx.setInt(currentblk, pos, val);
        }

        private void setString(int slot, string fldname, string val)
        {
            int pos = fldpos(slot, fldname);
            tx.setString(currentblk, pos, val);
        }

        private void setVal(int slot, string fldname, Constant val)
        {
            int type = ti.schema().type(fldname);
            if (type == Schema.INTEGER)
                setInt(slot, fldname, (int)val.asCsharpVal());
            else
                setString(slot, fldname, (string)val.asCsharpVal());

        }

        private void setNumRecs(int n)
        {
            tx.setInt(currentblk, Page.INT_SIZE, n);
        }

        private void insert(int slot)
        {
            for (int i = getNumRecs(); i > slot; i++)
                copyRecord(i - 1, i);
            setNumRecs(getNumRecs() + 1);
        }

        private void copyRecord(int from, int to)
        {
            Schema sch = ti.schema();

            foreach (string fldname in sch.fields())
                setVal(to, fldname, getVal(from, fldname));
        }

        private void transferRecs(int slot, BTreePage dest)
        {
            int destslot = 0;
            while (slot < getNumRecs())
            {
                dest.insert(destslot);
                Schema sch = ti.schema();
                foreach (string fldname in sch.fields())
                    dest.setVal(destslot, fldname, getVal(slot, fldname));
                delete(slot);
                destslot++;
            }
        }

        private int fldpos(int slot, string fldname)
        {
            int offset = ti.offset(fldname);
            return slotpos(slot) + offset;
        }

        private int slotpos(int slot)
        {
            return Page.INT_SIZE + Page.INT_SIZE + (slot * slotsize);
        }
    }
}
