﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.tx;
using simpledb.record;
using simpledb.metadata;
using simpledb.query;

namespace simpledb.index.query
{
    /// <summary>
    /// The Plan class corresponding to the <i>indexjoin</i>
    /// relational algebra operator.
    /// <author>Edward Sciore</author> 
    /// </summary>
    public class IndexJoinPlan : Plan
    {
        private Plan p1, p2;
        private IndexInfo ii;
        private string joinfield;
        private Schema sch = new Schema();

        /// <summary>
        /// Implements the join operator,
        /// using the specified LHS and RHS plans.
        /// </summary>
        /// <param name="p1">the left-hand plan</param>
        /// <param name="p2">the right-hand plan</param>
        /// <param name="ii">information about the right-hand index</param>
        /// <param name="joinfield">the left-hand field used for joining</param>
        /// <param name="tx">the calling transaction</param>
        public IndexJoinPlan(Plan p1, Plan p2, IndexInfo ii, string joinfield, Transaction tx)
        {
            this.p1 = p1;
            this.p2 = p2;
            this.ii = ii;
            this.joinfield = joinfield;
            sch.addAll(p1.schema());
            sch.addAll(p2.schema());
        }

        /// <summary>
        /// Opens an indexjoin scan for this query
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.open"/> 
        public Scan open()
        {
            Scan s = p1.open();
            TableScan ts = (TableScan)p2.open();
            Index idx = ii.open();
            return new IndexJoinScan(s, idx, joinfield, ts);
        }

        /// <summary>
        /// Estimates the number of block accesses to compute the join.
        /// The formula is:
        /// <code> B(indexjoin(p1,p2,idx)) = B(p1) + R(p1)*B(idx)
        ///       + R(indexjoin(p1,p2,idx) </code>
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.blocksAccessed"/>
        public int blocksAccessed()
        {
            return p1.blocksAccessed()
                    + (p1.recordsOutput() * ii.blocksAccessed())
                    + recordsOutput();
        }

        /// <summary>
        /// Estimates the number of output records in the join.
        /// The formula is:
        /// <code> R(indexjoin(p1,p2,idx)) = R(p1)*R(idx) </code>
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.recordsOutput"/>
        public int recordsOutput()
        {
            return p1.recordsOutput() * ii.recordsOutput();
        }

        /// <summary>
        /// Estimates the number of distinct values for the 
        /// specified field.
        /// </summary>  
        /// <seealso cref="simpledb.query.Plan.distinctValues"/>
        public int distinctValues(string fldname)
        {
            if (p1.schema().hasField(fldname))
                return p1.distinctValues(fldname);
            else
                return p2.distinctValues(fldname);
        }

        /// <summary>
        /// Returns the schema of the index join.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.schema"/>
        public Schema schema()
        {
            return sch;
        }
    }
}
