﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.query;

namespace simpledb.index.query
{
    /// <summary>
    /// The scan class corresponding to the indexjoin relational
    /// algebra operator.
    /// The code is very similar to that of ProductScan, 
    /// which makes sense because an index join is essentially
    /// the product of each LHS record with the matching RHS index records.
    /// </summary>
    /// <author>Edward Sciore</author> 
    public class IndexJoinScan : Scan
    {
        private Scan s;
        private TableScan ts; // the data table
        private Index idx;
        private string joinfield;

        /// <summary>
        /// Creates an index join scan for the specified LHS scan and
        /// RHS index.
        /// </summary>
        /// <param name="s">the LHS scan</param>
        /// <param name="idx">dx the RHS index</param>
        /// <param name="joinfield">the LHS field used for joining</param>
        /// <param name="ts">The tablescan.</param>
        public IndexJoinScan(Scan s, Index idx, string joinfield, TableScan ts)
        {
            this.s = s;
            this.idx = idx;
            this.joinfield = joinfield;
            this.ts = ts;
            beforeFirst();
        }

        /// <summary>
        /// Positions the scan before the first record.
        /// That is, the LHS scan will be positioned at its
        /// first record, and the index will be positioned
        /// before the first record for the join value.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/>  
        public void beforeFirst()
        {
            s.beforeFirst();
            s.next();
            resetIndex();
        }

        /// <summary>
        /// Moves the scan to the next record.
        /// The method moves to the next index record, if possible.
        /// Otherwise, it moves to the next LHS record and the
        /// first index record.
        /// If there are no more LHS records, the method returns false.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/>
        public bool next()
        {
            while (true)
            {
                if (idx.next())
                {
                    ts.moveToRid(idx.getDataRid());
                    return true;
                }
                if (!s.next())
                    return false;
                resetIndex();
            }
        }

        /// <summary>
        /// Closes the scan by closing its LHS scan and its RHS index.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close"/>  
        public void close()
        {
            s.close();
            idx.close();
            ts.close();
        }

        /// <summary>
        /// Returns the Constant value of the specified field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public Constant getVal(string fldname)
        {
            if (ts.hasField(fldname))
                return ts.getVal(fldname);
            else
                return s.getVal(fldname);
        }

        /// <summary>
        /// Returns the integer value of the specified field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public int getInt(string fldname)
        {
            if (ts.hasField(fldname))
                return ts.getInt(fldname);
            else
                return s.getInt(fldname);
        }

        /// <summary>
        /// Returns the string value of the specified field.
        /// <seealso cref="simpledb.query.Scan.getVal"/>  
        /// </summary>
        /// <param name="fldname"></param>
        /// <returns></returns>
        public string getString(string fldname)
        {
            if (ts.hasField(fldname))
                return ts.getString(fldname);
            else
                return s.getString(fldname);
        }

        /// <summary>
        /// Returns true if the field is in the schema.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/> 
        public bool hasField(string fldname)
        {
            return ts.hasField(fldname) || s.hasField(fldname);
        }

        private void resetIndex()
        {
            Constant searchkey = s.getVal(joinfield);
            idx.beforeFirst(searchkey);
        }
    }
}
