﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.record;
using simpledb.query;

namespace simpledb.index.query
{
    /// <summary>
    /// The scan class corresponding to the select relational
    /// algebra operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class IndexSelectScan : Scan
    {
        private Index idx;
        private Constant val;
        private TableScan ts;

        /// <summary>
        /// Creates an index select scan for the specified
        /// index and selection constant.
        /// </summary>
        /// <param name="idx">dx the index</param>
        /// <param name="val">the selection constant</param>
        /// <param name="ts">The tablescan</param>
        public IndexSelectScan(Index idx, Constant val, TableScan ts)
        {
            this.idx = idx;
            this.val = val;
            this.ts = ts;
            beforeFirst();
        }

        /// <summary>
        /// Positions the scan before the first record,
        /// which in this case means positioning the index
        /// before the first instance of the selection constant.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/>
        public void beforeFirst()
        {
            idx.beforeFirst(val);
        }

        /// <summary>
        /// Moves to the next record, which in this case means
        /// moving the index to the next record satisfying the
        /// selection constant, and returning false if there are
        /// no more such index records.
        /// If there is a next record, the method moves the 
        /// tablescan to the corresponding data record.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/>
        public bool next()
        {
            bool ok = idx.next();
            if (ok)
            {
                RID rid = idx.getDataRid();
                ts.moveToRid(rid);
            }
            return ok;
        }

        /// <summary>
        /// Closes the scan by closing the index and the tablescan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close"/>  
        public void close()
        {
            idx.close();
            ts.close();
        }

        /// <summary>
        /// Returns the value of the field of the current data record.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/>
        public Constant getVal(string fldname)
        {
            return ts.getVal(fldname);
        }

        /// <summary>
        /// Returns the value of the field of the current data record.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getInt"/>
        public int getInt(string fldname)
        {
            return ts.getInt(fldname);
        }

        /// <summary>
        /// Returns the value of the field of the current data record.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getString"/>
        public string getString(string fldname)
        {
            return ts.getString(fldname);
        }

        /// <summary>
        /// Returns whether the data record has the specified field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/> 
        public bool hasField(string fldname)
        {
            return ts.hasField(fldname);
        }
    }
}
