// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections;
using System.Collections.Generic;
using simpledb.file;
using simpledb.server;

namespace simpledb.log
{
    /// <summary>
    /// The low-level log manager.
    /// This log manager is responsible for writing log records
    /// into a log file.
    /// A log record can be any sequence of integer and string values.
    /// The log manager does not understand the meaning of these
    /// values, which are written and read by the
    /// <see cref="simpledb.tx.recovery.RecoveryMgr"/>.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class LogMgr : IEnumerable<BasicLogRecord>
    {
        /// <summary>
        /// The location where the pointer to the last integer in the page is.
        /// A value of 0 means that the pointer is the first value in the page.
        /// </summary>
        internal const int LAST_POS = 0;

        private string logfile;
        private Page mypage = new Page();
        private Block currentblk;
        private int currentpos;
        private object threadLock = new object();

        /// <summary>
        /// Creates the manager for the specified log file.
        /// If the log file does not yet exist, it is created
        /// with an empty first block.
        /// This constructor depends on a <see cref="FileMgr" /> object
        /// that it gets from the method
        /// <see cref="simpledb.server.SimpleDB.fileMgr"/>.
        /// That object is created during system initialization.
        /// Thus this constructor cannot be called until
        /// <see cref="simpledb.server.SimpleDB.initFileMgr"/> 
        /// is called first.
        /// </summary>
        /// <param name="logfile">the name of the log file</param>
        public LogMgr(string logfile)
        {
            this.logfile = logfile;
            int logsize = SimpleDB.fileMgr().size(logfile);
            if (logsize == 0)
                appendNewBlock();
            else
            {
                currentblk = new Block(logfile, logsize - 1);
                mypage.read(currentblk);
                currentpos = getLastRecordPosition() + Page.INT_SIZE;
            }
        }

        /// <summary>
        /// Ensures that the log records corresponding to the
        /// specified LSN has been written to disk.
        /// All earlier log records will also be written to disk.
        /// </summary>
        /// <param name="lsn">the LSN of a log record</param>
        public void flush(int lsn)
        {
            if (lsn >= currentLSN())
                flush();
        }

        /// <summary>
        ///  Returns an iterator for the log records,
        /// which will be returned in reverse order starting with the most recent.
        /// </summary>
        public IEnumerator<BasicLogRecord> GetEnumerator()
        {
            lock (threadLock)
            {
                flush();
                return new LogIterator(currentblk);
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            lock (threadLock)
            {
                flush();
                return new LogIterator(currentblk);
            }
        }

        /// <summary>
        /// Appends a log record to the file.
        /// The record contains an arbitrary array of strings and integers.
        /// The method also writes an integer to the end of each log record whose value
        /// is the offset of the corresponding integer for the previous log record.
        /// These integers allow log records to be read in reverse order.
        /// </summary>
        /// <param name="rec">the list of values</param>  
        /// <returns>the LSN of the final value</returns>      
        public int append(object[] rec)
        {
            lock (threadLock)
            {
                int recsize = Page.INT_SIZE;

                foreach (object obj in rec)
                    recsize += size(obj);
                if (currentpos + recsize >= Page.BLOCK_SIZE)
                {
                    flush();
                    appendNewBlock();
                }
                foreach (object obj in rec)
                    appendVal(obj);
                finalizeRecord();
                return currentLSN();
            }
        }

        /// <summary>
        /// Adds the specified value to the page at the position denoted by
        /// currentpos. Then increments currentpos by the size of the value.
        /// </summary>
        /// <param name="val">the integer or string to be added to the page</param>      
        private void appendVal(object val)
        {
            if (val is string)
                mypage.setString(currentpos, (string)val);
            else
                mypage.setInt(currentpos, (int)val);
            currentpos += size(val);

        }

        /// <summary>
        ///  Calculates the size of the specified integer or string.
        /// </summary> 
        /// <param name="val">the value</param> 
        /// <returns>the size of the value, in bytes</returns> 
        private int size(object val)
        {
            if (val is string)
            {
                string sval = (string)val;
                return Page.STR_SIZE(sval.Length);
            }
            else
                return Page.INT_SIZE;
        }

        /// <summary>
        /// Returns the LSN of the most recent log record.
        /// As implemented, the LSN is the block number where the record is stored.
        /// Thus every log record in a block has the same LSN.
        /// </summary>     
        /// <returns>the LSN of the most recent log record</returns> 
        private int currentLSN()
        {
            return currentblk.number();
        }

        /// <summary>
        ///  Writes the current page to the log file.
        /// </summary>
        private void flush()
        {
            mypage.write(currentblk);
        }

        /// <summary>
        ///  Clear the current page, and append it to the log file.
        /// </summary>
        private void appendNewBlock()
        {
            setLastRecordPosition(0);
            currentpos = Page.INT_SIZE;
            currentblk = mypage.append(logfile);
        }

        /// <summary>
        /// Sets up a circular chain of pointers to the records in the page.
        /// There is an integer added to the end of each log record
        /// whose value is the offset of the previous log record.
        /// The first four bytes of the page contain an integer whose value
        /// is the offset of the integer for the last log record in the page.
        /// </summary>
        private void finalizeRecord()
        {
            mypage.setInt(currentpos, getLastRecordPosition());
            setLastRecordPosition(currentpos);
            currentpos += Page.INT_SIZE;
        }

        private int getLastRecordPosition()
        {
            return mypage.getInt(LAST_POS);
        }

        private void setLastRecordPosition(int pos)
        {
            mypage.setInt(LAST_POS, pos);
        }
    }
}
