﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.query;

namespace simpledb.materialize
{
    /// <summary>
    /// The Scan class for the <i>groupby</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    class GroupByScan : Scan
    {
        private Scan s;
        private List<string> groupfields;
        private List<AggregationFn> aggfns;
        private GroupValue groupval;
        private bool moregroups;

        /// <summary>
        ///  Creates a groupby scan, given a grouped table scan.
        /// </summary>
        /// <param name="s">the grouped scan</param> 
        /// <param name="groupfields">the group fields</param>  
        /// <param name="aggfns">the aggregation functions</param> 
        public GroupByScan(Scan s, List<string> groupfields, List<AggregationFn> aggfns)
        {
            this.s = s;
            this.groupfields = groupfields;
            this.aggfns = aggfns;
            beforeFirst();
        }

        /// <summary>
        /// Positions the scan before the first group.
        /// Internally, the underlying scan is always 
        /// positioned at the first record of a group, which 
        /// means that this method moves to the
        /// first underlying record.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/>
        public void beforeFirst()
        {
            s.beforeFirst();
            moregroups = s.next();
        }

        /// <summary>
        /// Moves to the next group.
        /// The key of the group is determined by the 
        /// group values at the current record.
        /// The method repeatedly reads underlying records until
        /// it encounters a record having a different key.
        /// The aggregation functions are called for each record
        /// in the group. 
        /// The values of the grouping fields for the group are saved.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/> 
        public bool next()
        {
            if (!moregroups)
                return false;
            foreach (AggregationFn fn in aggfns)
                fn.processFirst(s);
            groupval = new GroupValue(s, groupfields);
            while (moregroups = s.next())
            {
                GroupValue gv = new GroupValue(s, groupfields);
                if (!groupval.Equals(gv))
                    break;
                foreach (AggregationFn fn in aggfns)
                    fn.processNext(s);
            }
            return true;
        }

        /// <summary>
        /// Closes the scan by closing the underlying scan.       
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close"/> 
        public void close()
        {
            s.close();
        }

        /// <summary>
        /// Gets the Constant value of the specified field.
        /// If the field is a group field, then its value can
        /// be obtained from the saved group value.
        /// Otherwise, the value is obtained from the
        /// appropriate aggregation function.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public Constant getVal(string fldname)
        {
            if (groupfields.Contains(fldname))
                return groupval.getVal(fldname);
            foreach (AggregationFn fn in aggfns)
                if (fn.fieldName().Equals(fldname))
                    return fn.value();
            throw new System.Exception("field" + fldname + "not found.");
        }

        /// <summary>
        /// Gets the integer value of the specified field.
        /// If the field is a group field, then its value can
        /// be obtained from the saved group value.
        /// Otherwise, the value is obtained from the
        /// appropriate aggregation function.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public int getInt(string fldname)
        {
            return (int)getVal(fldname).asCsharpVal();
        }

        /// <summary>
        ///  Gets the string value of the specified field.
        /// If the field is a group field, then its value can
        /// be obtained from the saved group value.
        /// Otherwise, the value is obtained from the
        /// appropriate aggregation function.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public string getString(string fldname)
        {
            return (string)getVal(fldname).asCsharpVal();
        }

        /// <summary>
        /// Returns true if the specified field is either a 
        /// grouping field or created by an aggregation function.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/>
        public bool hasField(string fldname)
        {
            if (groupfields.Contains(fldname))
                return true;
            foreach (AggregationFn fn in aggfns)
                if (fn.fieldName().Equals(fldname))
                    return true;
            return false;
        }
    }
}
