﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.tx;
using simpledb.file;
using simpledb.record;
using simpledb.query;

namespace simpledb.materialize
{
    /// <summary>
    ///  The Plan class for the <i>materialize</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class MaterializePlan : Plan
    {
        private Plan srcplan;
        private Transaction tx;

        /// <summary>
        ///  Creates a materialize plan for the specified query.
        /// </summary>
        /// <param name="srcplan">the plan of the underlying query</param>  
        /// <param name="tx">the calling transaction</param>  
        public MaterializePlan(Plan srcplan, Transaction tx)
        {
            this.srcplan = srcplan;
            this.tx = tx;
        }

        /// <summary>
        ///  This method loops through the underlying query,
        /// copying its output records into a temporary table.
        /// It then returns a table scan for that table.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.open"/> 
        public Scan open()
        {
            Schema sch = srcplan.schema();
            TempTable temp = new TempTable(sch, tx);
            Scan src = srcplan.open();
            UpdateScan dest = temp.open();
            while (src.next())
            {
                dest.insert();
                foreach (string fldname in sch.fields())
                    dest.setVal(fldname, src.getVal(fldname));
            }
            src.close();
            dest.beforeFirst();
            return dest;
        }

        /// <summary>
        ///  Returns the estimated number of blocks in the 
        /// materialized table.
        /// It does <i>not</i> include the one-time cost
        /// of materializing the records.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.blocksAccessed"/> 
        public int blocksAccessed()
        {
            TableInfo ti = new TableInfo("", srcplan.schema());
            double rpb = (double)(Page.BLOCK_SIZE / ti.recordLength());
            return (int)System.Math.Ceiling(srcplan.recordsOutput() / rpb);
        }

        /// <summary>
        ///  Returns the number of records in the materialized table,
        /// which is the same as in the underlying plan.       
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.recordsOutput"/> 
        public int recordsOutput()
        {
            return srcplan.recordsOutput();
        }

        /// <summary>
        ///  Returns the number of distinct field values,
        /// which is the same as in the underlying plan.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.distinctValues"/> 
        public int distinctValues(string fldname)
        {
            return srcplan.distinctValues(fldname);
        }

        /// <summary>
        ///  Returns the schema of the materialized table,
        /// which is the same as in the underlying plan.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.schema"/> 
        public Schema schema()
        {
            return srcplan.schema();
        }
    }
}
