﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.tx;
using simpledb.record;
using simpledb.query;


namespace simpledb.materialize
{
    /// <summary>
    ///  The Plan class for the <i>mergejoin</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class MergeJoinPlan : Plan
    {
        private Plan p1, p2;
        private string fldname1, fldname2;
        private Schema sch = new Schema();

        /// <summary>
        /// Creates a mergejoin plan for the two specified queries.
        /// The RHS must be materialized after it is sorted, 
        /// in order to deal with possible duplicates.
        /// </summary>
        /// <param name="p1">the LHS query plan</param>  
        /// <param name="p2">the RHS query plan</param>  
        /// <param name="fldname1"> the LHS join field</param> 
        /// <param name="fldname2">the RHS join field</param>
        /// <param name="tx">the calling transaction</param>  
        public MergeJoinPlan(Plan p1, Plan p2, string fldname1, string fldname2, Transaction tx)
        {
            this.fldname1 = fldname1;
            List<string> sortlist1 = new List<string>();
            sortlist1.Add(fldname1);
            this.p1 = new SortPlan(p1, sortlist1, tx);

            this.fldname2 = fldname2;
            List<string> sortlist2 = new List<string>();
            sortlist2.Add(fldname2);
            this.p2 = new SortPlan(p2, sortlist2, tx);

            sch.addAll(p1.schema());
            sch.addAll(p2.schema());
        }

        /// <summary>
        ///  The method first sorts its two underlying scans
        /// on their join field. It then returns a mergejoin scan
        /// of the two sorted table scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.open"/> 
        public Scan open()
        {
            Scan s1 = p1.open();
            SortScan s2 = (SortScan)p2.open();
            return new MergeJoinScan(s1, s2, fldname1, fldname2);
        }

        /// <summary>
        ///  Returns the number of block acceses required to
        /// mergejoin the sorted tables.
        /// Since a mergejoin can be preformed with a single
        /// pass through each table, the method returns
        /// the sum of the block accesses of the 
        /// materialized sorted tables.
        /// It does <i>not</i> include the one-time cost
        /// of materializing and sorting the records.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.blocksAccessed"/> 
        public int blocksAccessed()
        {
            return p1.blocksAccessed() + p2.blocksAccessed();
        }

        /// <summary>
        ///  Returns the number of records in the join.
        ///  Assuming uniform distribution, the formula is:
        /// <code> R(join(p1,p2)) = R(p1)*R(p2)/max{V(p1,F1),V(p2,F2)}</code>
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.recordsOutput"/>
        public int recordsOutput()
        {
            int maxvals = System.Math.Max(p1.distinctValues(fldname1), p2.distinctValues(fldname2));
            return (p1.recordsOutput() * p2.recordsOutput()) / maxvals;
        }

        /// <summary>
        ///  Estimates the distinct number of field values in the join.
        /// Since the join does not increase or decrease field values,
        /// the estimate is the same as in the appropriate underlying query.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.distinctValues"/> 
        public int distinctValues(string fldname)
        {
            if (p1.schema().hasField(fldname))
                return p1.distinctValues(fldname);
            else
                return p2.distinctValues(fldname);
        }

        /// <summary>
        /// Returns the schema of the join,
        /// which is the union of the schemas of the underlying queries.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.schema"/>
        public Schema schema()
        {
            return sch;
        }
    }
}
