﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.query;

namespace simpledb.materialize
{
    /// <summary>
    /// The Scan class for the <i>mergejoin</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    class MergeJoinScan : Scan
    {
        private Scan s1;
        private SortScan s2;
        private string fldname1, fldname2;
        private Constant joinval = null;

        /// <summary>
        ///  Creates a mergejoin scan for the two underlying sorted scans.
        /// </summary>
        /// <param name="s1">the LHS sorted scan</param>  
        /// <param name="s2">the RHS sorted scan</param>  
        /// <param name="fldname1"> the LHS join field</param> 
        /// <param name="fldname2">f the RHS join field</param> 
        public MergeJoinScan(Scan s1, SortScan s2, string fldname1, string fldname2)
        {
            this.s1 = s1;
            this.s2 = s2;
            this.fldname1 = fldname1;
            this.fldname2 = fldname2;
            beforeFirst();
        }

        /// <summary>
        ///  Positions the scan before the first record,
        /// by positioning each underlying scan before
        /// their first records.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/> 
        public void beforeFirst()
        {
            s1.beforeFirst();
            s2.beforeFirst();
        }

        /// <summary>
        ///  Closes the scan by closing the two underlying scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close()"/> 
        public void close()
        {
            s1.close();
            s2.close();
        }

        /// <summary>
        ///  Moves to the next record.  This is where the action is.
        /// If the next RHS record has the same join value,
        /// then move to it.
        /// Otherwise, if the next LHS record has the same join value,
        /// then reposition the RHS scan back to the first record
        /// having that join value.
        /// Otherwise, repeatedly move the scan having the smallest
        /// value until a common join value is found.
        /// When one of the scans runs out of records, return false.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/> 
        public bool next()
        {
            bool hasmore2 = s2.next();
            if (hasmore2 && s2.getVal(fldname2).Equals(joinval))
                return true;

            bool hasmore1 = s1.next();
            if (hasmore1 && s1.getVal(fldname1).Equals(joinval))
            {
                s2.restorePosition();
                return true;
            }

            while (hasmore1 && hasmore2)
            {
                Constant v1 = s1.getVal(fldname1);
                Constant v2 = s2.getVal(fldname2);
                if (v1.CompareTo(v2) < 0)
                    hasmore1 = s1.next();
                else if (v1.CompareTo(v2) > 0)
                    hasmore2 = s2.next();
                else
                {
                    s2.savePosition();
                    joinval = s2.getVal(fldname2);
                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Returns the value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/>
        public Constant getVal(string fldname)
        {
            if (s1.hasField(fldname))
                return s1.getVal(fldname);
            else
                return s2.getVal(fldname);
        }

        /// <summary>
        ///  Returns the integer value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getInt"/> 
        public int getInt(string fldname)
        {
            if (s1.hasField(fldname))
                return s1.getInt(fldname);
            else
                return s2.getInt(fldname);
        }

        /// <summary>
        ///  Returns the string value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getString"/>
        public string getString(string fldname)
        {
            if (s1.hasField(fldname))
                return s1.getString(fldname);
            else
                return s2.getString(fldname);
        }

        /// <summary>
        /// Returns true if the specified field is in
        /// either of the underlying scans.
        /// </summary>
        /// <seealso cref=" simpledb.query.Scan.hasField"/>
        public bool hasField(string fldname)
        {
            return s1.hasField(fldname) || s2.hasField(fldname);
        }
    }
}
