﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.tx;
using simpledb.record;
using simpledb.query;

namespace simpledb.materialize
{
    /// <summary>
    /// The Plan class for the <i>sort</i> operator. 
    /// </summary>
    ///  <author>Edward Sciore</author>
    class SortPlan : Plan
    {
        private Plan p;
        private Transaction tx;
        public Schema sch;
        private RecordComparator comp;

        /// <summary>
        /// Creates a sort plan for the specified query.
        /// </summary>
        /// <param name="p">the plan for the underlying query</param>  
        /// <param name="sortfields">the fields to sort by</param> 
        /// <param name="tx">the calling transaction</param>  
        public SortPlan(Plan p, List<string> sortfields, Transaction tx)
        {
            this.p = p;
            this.tx = tx;
            sch = p.schema();
            comp = new RecordComparator(sortfields);
        }

        /// <summary>
        /// This method is where most of the action is.
        /// Up to 2 sorted temporary tables are created,
        /// and are passed into SortScan for final merging.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.open"/> 
        public Scan open()
        {
            Scan src = p.open();
            List<TempTable> runs = splitIntoRuns(src);
            src.close();
            while (runs.Count > 2)
                runs = mergeAllRuns(runs);
            return (Scan)new SortScan(runs, comp);
        }

        /// <summary>
        ///  Returns the number of blocks in the sorted table,
        /// which is the same as it would be in a
        /// materialized table.
        /// It does <i>not</i> include the one-time cost
        /// of materializing and sorting the records.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.blocksAccessed"/> 
        public int blocksAccessed()
        {
            // does not include the one-time cost of sorting
            Plan mp = new MaterializePlan(p, tx); // not opened; just for analysis
            return mp.blocksAccessed();
        }

        /// <summary>
        ///  Returns the number of records in the sorted table,
        /// which is the same as in the underlying query.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.recordsOutput"/> 
        public int recordsOutput()
        {
            return p.recordsOutput();
        }

        /// <summary>
        /// Returns the number of distinct field values in
        /// the sorted table, which is the same as in
        /// the underlying query.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.distinctValues"/> 
        public int distinctValues(string fldname)
        {
            return p.distinctValues(fldname);
        }

        /// <summary>
        ///  Returns the schema of the sorted table, which
        /// is the same as in the underlying query.
        /// </summary>
        /// <seealso cref="simpledb.query.Plan.schema"/> 
        public Schema schema()
        {
            return sch;
        }

        private List<TempTable> splitIntoRuns(Scan src)
        {
            List<TempTable> temps = new List<TempTable>();
            src.beforeFirst();
            if (!src.next())
                return temps;
            TempTable currenttemp = new TempTable(sch, tx);
            temps.Add(currenttemp);
            UpdateScan currentscan = currenttemp.open();
            while (copy(src, currentscan))
                if (comp.compare(src, currentscan) < 0)
                {
                    // start a new run
                    currentscan.close();
                    currenttemp = new TempTable(sch, tx);
                    temps.Add(currenttemp);
                    currentscan = (UpdateScan)currenttemp.open();
                }
            currentscan.close();
            return temps;
        }

        private List<TempTable> mergeAllRuns(List<TempTable> runs)
        {
            List<TempTable> result = new List<TempTable>();
            while (runs.Count > 1)
            {

                TempTable p1 = runs[0];  
                runs.RemoveAt(0);
                TempTable p2 = runs[0];
                runs.RemoveAt(0);

                result.Add(mergeTwoRuns(p1, p2));
            }
            if (runs.Count == 1)
                result.Add(runs[0]);
            return result;
        }

        private TempTable mergeTwoRuns(TempTable p1, TempTable p2)
        {
            Scan src1 = p1.open();
            Scan src2 = p2.open();
            TempTable result = new TempTable(sch, tx);
            UpdateScan dest = result.open();

            bool hasmore1 = src1.next();
            bool hasmore2 = src2.next();
            while (hasmore1 && hasmore2)

                if (comp.compare(src1, src2) < 0)
                    hasmore1 = copy(src1, dest);
                else
                    hasmore2 = copy(src2, dest);

            if (hasmore1)
                while (hasmore1)
                    hasmore1 = copy(src1, dest);
            else
                while (hasmore2)
                    hasmore2 = copy(src2, dest);

            src1.close();
            src2.close();
            dest.close();
            return result;

        }

        private bool copy(Scan src, UpdateScan dest)
        {
            dest.insert();
            foreach (string fldname in sch.fields())
                dest.setVal(fldname, src.getVal(fldname));
            return src.next();
        }
    }
}
