﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.record;
using simpledb.query;

namespace simpledb.materialize
{
    /// <summary>
    ///  The Scan class for the <i>sort</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    class SortScan : Scan
    {
        private UpdateScan s1, s2 = null, currentscan = null;
        private RecordComparator comp;
        private bool hasmore1, hasmore2 = false;
        private List<RID> saveposition = null;

        /// <summary>
        /// Creates a sort scan, given a list of 1 or 2 runs.
        /// If there is only 1 run, then s2 will be null and
        /// hasmore2 will be false.
        /// </summary>
        /// <param name="runs">the list of runs</param>  
        /// <param name="comp">the record comparator</param> 
        public SortScan(List<TempTable> runs, RecordComparator comp)
        {
            this.comp = comp;
            s1 = (UpdateScan)runs[0].open();
            hasmore1 = s1.next();
            if (runs.Count > 1)
            {
                s2 = (UpdateScan)runs[1].open();
                hasmore2 = s2.next();
            }
        }
        /// <summary>
        /// Positions the scan before the first record in sorted order.
        /// Internally, it moves to the first record of each underlying scan.
        /// The variable currentscan is set to null, indicating that there is
        /// no current scan.
        /// </summary> 
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/> 
        public void beforeFirst()
        {
            s1.beforeFirst();
            hasmore1 = s1.next();
            if (s2 != null)
            {
                s2.beforeFirst();
                hasmore2 = s2.next();
            }
        }

        /// <summary>
        /// Moves to the next record in sorted order.
        /// First, the current scan is moved to the next record.
        /// Then the lowest record of the two scans is found, and that
        /// scan is chosen to be the new current scan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/> 
        public bool next()
        {
            if (currentscan == s1)
                hasmore1 = s1.next();
            else if (currentscan == s2)
                hasmore2 = s2.next();

            if (!hasmore1 && !hasmore2)
                return false;
            else if (hasmore1 && hasmore2)
            {
                if (comp.compare(s1, s2) < 0)
                    currentscan = s1;
                else
                    currentscan = s2;
            }
            else if (hasmore1)
                currentscan = s1;
            else if (hasmore2)
                currentscan = s2;
            return true;
        }

        /// <summary>
        ///  Closes the two underlying scans.
        /// </summary> 
        /// <seealso cref="simpledb.query.Scan.close"/> 
        public void close()
        {
            s1.close();
            if (s2 != null)
                s2.close();
        }

        /// <summary>
        /// Gets the Constant value of the specified field
        /// of the current scan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/> 
        public Constant getVal(string fldname)
        {
            return currentscan.getVal(fldname);
        }

        /// <summary>
        /// Gets the integer value of the specified field
        /// of the current scan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getInt"/> 
        public int getInt(string fldname)
        {
            return currentscan.getInt(fldname);
        }

        /// <summary>
        /// Gets the string value of the specified field
        /// of the current scan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getString"/> 
        public string getString(string fldname)
        {
            return currentscan.getString(fldname);
        }

        /// <summary>
        ///  Returns true if the specified field is in the current scan.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/>
        public bool hasField(string fldname)
        {
            return currentscan.hasField(fldname);
        }

        /// <summary>
        /// Saves the position of the current record,
        /// so that it can be restored at a later time.
        /// </summary>
        public void savePosition()
        {
            RID rid1 = s1.getRid();
            RID rid2 = s2.getRid();
            saveposition = new List<RID>();
            saveposition.Add(rid1);
            saveposition.Add(rid2);
        }

        /// <summary>
        ///  Moves the scan to its previously-saved position.
        /// </summary>
        public void restorePosition()
        {
            RID rid1 = saveposition[0];
            RID rid2 = saveposition[1];
            s1.moveToRid(rid1);
            s2.moveToRid(rid2);
        }
    }
}
