// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.record;
using simpledb.tx;
using simpledb.server;
using simpledb.index;
using simpledb.index.hash;

namespace simpledb.metadata
{
    /// <summary>
    /// The information about an index.
    /// This information is used by the query planner in order to
    /// estimate the costs of using the index,
    /// and to obtain the schema of the index records.
    /// It's methods are essentially the same as those of Plan.
    /// </summary>
    /// <author>Edward Sciore</author> 
    public class IndexInfo
    {
        private string idxname, fldname;
        private Transaction tx;
        private TableInfo ti;
        private StatInfo si;

        /// <summary>
        /// Creates an IndexInfo object for the specified index.
        /// </summary>
        /// <param name="idxname">the name of the index</param>
        /// <param name="tblname">the name of the table</param>
        /// <param name="fldname">the name of the indexed field</param>
        /// <param name="tx">the calling transaction</param>
        public IndexInfo(string idxname, string tblname, string fldname, Transaction tx)
        {
            this.idxname = idxname;
            this.fldname = fldname;
            this.tx = tx;
            ti = SimpleDB.mdMgr().getTableInfo(tblname, tx);
            si = SimpleDB.mdMgr().getStatInfo(tblname, ti, tx);
        }

        /// <summary>
        /// Opens the index described by this object.
        /// </summary>
        /// <returns>the Index object associated with this information</returns>
        public Index open()
        {
            Schema sch = schema();
            return new HashIndex(idxname, sch, tx);
        }

        /// <summary>
        /// Estimates the number of block accesses required to
        /// find all index records having a particular search key.
        /// The method uses the table's metadata to estimate the
        /// size of the index file and the number of index records
        /// per block.
        /// It then passes this information to the searchCost
        /// method of the appropriate index type,
        /// which provides the estimate.
        /// </summary>
        /// <returns>the number of block accesses required to traverse the index</returns>
        public int blocksAccessed()
        {
            int rpb = Page.BLOCK_SIZE / ti.recordLength();
            int numblocks = si.recordsOutput() / rpb;
            return HashIndex.searchCost(numblocks, rpb);
        }

        /// <summary>
        /// Returns the estimated number of records having a
        /// search key.  This value is the same as doing a select
        /// query; that is, it is the number of records in the table
        /// divided by the number of distinct values of the indexed field.
        /// </summary>
        /// <returns>the estimated number of records having a search key</returns>
        public int recordsOutput()
        {
            return si.recordsOutput() / si.distinctValues(fldname);
        }

        /// <summary>
        /// Returns the distinct values for a specified field
        /// in the underlying table, or 1 for the indexed field.
        /// </summary>
        /// <param name="fname">the specified field</param>
        /// <returns>The distinct values for a specified field</returns>
        public int distinctValues(string fname)
        {
            if (fldname.Equals(fldname))
                return 1;
            else
                return System.Math.Min(si.distinctValues(fldname), recordsOutput());
        }

        /// <summary>
        /// Returns the schema of the index records.
        /// The schema consists of the dataRID (which is
        /// represented as two integers, the block number and the
        /// record ID) and the dataval (which is the indexed field).
        /// Schema information about the indexed field is obtained
        /// via the table's metadata.
        /// </summary>
        /// <returns>the schema of the index records</returns>
        private Schema schema()
        {
            Schema sch = new Schema();
            sch.addIntField("block");
            sch.addIntField("id");   
            if (ti.schema().type(fldname) == Schema.INTEGER)
                sch.addIntField("dataval");
            else
            {
                int fldlen = ti.schema().length(fldname);
                sch.addStringField("dataval", fldlen);
            }
            return sch;
        }
    }
}
