// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.tx;
using simpledb.record;

namespace simpledb.metadata
{
    /// <summary>
    /// The table manager.
    /// There are methods to create a table, save the metadata
    /// in the catalog, and obtain the metadata of a
    /// previously-created table. 
    /// </summary>
    /// <author>Edward Sciore</author> 
    class TableMgr
    {
        /// <summary>
        /// The maximum number of characters in any
        /// tablename or fieldname.
        /// Currently, this value is 16.
        /// </summary>
        public const int MAX_NAME = 16;

        private TableInfo tcatInfo, fcatInfo;

        /// <summary>
        /// Creates a new catalog manager for the database system.
        /// If the database is new, then the two catalog tables
        /// are created.
        /// </summary>
        /// <param name="isNew">has the value true if the database is new</param>
        /// <param name="tx">the startup transaction</param>
        public TableMgr(bool isNew, Transaction tx)
        {
            Schema tcatSchema = new Schema();
            tcatSchema.addStringField("tblname", MAX_NAME);
            tcatSchema.addIntField("reclength");
            tcatInfo = new TableInfo("tblcat", tcatSchema);

            Schema fcatSchema = new Schema();
            fcatSchema.addStringField("tblname", MAX_NAME);
            fcatSchema.addStringField("fldname", MAX_NAME);
            fcatSchema.addIntField("type");
            fcatSchema.addIntField("length");
            fcatSchema.addIntField("offset");
            fcatInfo = new TableInfo("fldcat", fcatSchema);

            if (isNew)
            {
                createTable("tblcat", tcatSchema, tx);
                createTable("fldcat", fcatSchema, tx);
            }
        }

        /// <summary>
        /// Creates a new table having the specified name and schema.
        /// </summary>
        /// <param name="tblname">the name of the new table</param>
        /// <param name="sch">the table's schema</param>
        /// <param name="tx">the transaction creating the table</param>
        public void createTable(string tblname, Schema sch, Transaction tx)
        {
            TableInfo ti = new TableInfo(tblname, sch);

            // insert one record into tblcat
            RecordFile tcatfile = new RecordFile(tcatInfo, tx);
            tcatfile.insert();
            tcatfile.setString("tblname", tblname);
            tcatfile.setInt("reclength", ti.recordLength());
            tcatfile.close();

            // insert a record into fldcat for each field
            RecordFile fcatfile = new RecordFile(fcatInfo, tx);
            foreach (string fldname in sch.fields())
            {
                fcatfile.insert();
                fcatfile.setString("tblname", tblname);
                fcatfile.setString("fldname", fldname);
                fcatfile.setInt("type", sch.type(fldname));
                fcatfile.setInt("length", sch.length(fldname));
                fcatfile.setInt("offset", ti.offset(fldname));
            }
            fcatfile.close();
        }

        /// <summary>
        /// Retrieves the metadata for the specified table
        /// out of the catalog.
        /// </summary>
        /// <param name="tblname">the name of the table</param>
        /// <param name="tx">the transaction</param>
        /// <returns>the table's stored metadata</returns>
        public TableInfo getTableInfo(string tblname, Transaction tx)
        {
            RecordFile tcatfile = new RecordFile(tcatInfo, tx);
            int reclen = -1;
            while (tcatfile.next())
            {
                if (tcatfile.getString("tblname").Equals(tblname))
                {
                    reclen = tcatfile.getInt("reclength");
                    break;
                }
            }
            tcatfile.close();

            RecordFile fcatfile = new RecordFile(fcatInfo, tx);
            Schema sch = new Schema();
            Dictionary<string, int> offsets = new Dictionary<string, int>();
            while (fcatfile.next())
            {
                if (fcatfile.getString("tblname").Equals(tblname))
                {
                    string fldname = fcatfile.getString("fldname");
                    int fldtype = fcatfile.getInt("type");
                    int fldlen = fcatfile.getInt("length");
                    int offset = fcatfile.getInt("offset");
                    offsets.Add(fldname, offset);
                    sch.addFiled(fldname, fldtype, fldlen);
                }
            }
            fcatfile.close();
            return new TableInfo(tblname, sch, offsets, reclen);
        }
    }
}
