﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.tx;
using simpledb.record;
using simpledb.file;
using simpledb.query;

namespace simpledb.multibuffer
{
    /// <summary>
    /// The class for the <i>chunk</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class ChunkScan : Scan
    {
        private List<RecordPage> pages;
        private int startbnum, endbnum, current;
        private Schema sch;
        private RecordPage rp;

        /// <summary>
        /// Creates a chunk consisting of the specified pages. 
        /// </summary>
        /// <param name="ti">the metadata for the chunked table</param>
        /// <param name="startbnum">the starting block number</param>
        /// <param name="endbnum">the ending block number</param>
        /// <param name="tx">the current transaction</param>
        public ChunkScan(TableInfo ti, int startbnum, int endbnum, Transaction tx)
        {
            pages = new List<RecordPage>();
            this.startbnum = startbnum;
            this.endbnum = endbnum;
            this.sch = ti.schema();
            string filename = ti.fileName();
            for (int i = startbnum; i <= endbnum; i++)
            {
                Block blk = new Block(filename, i);
                pages.Add(new RecordPage(blk, ti, tx));
            }
            beforeFirst();
        }

        /// <summary>
        /// Positions the scan before the first record.
        /// </summary>
        ///  <seealso cref="simpledb.query.Scan.beforeFirst"/> 
        public void beforeFirst()
        {
            moveToBlock(startbnum);
        }

        /// <summary>
        /// Moves to the next record in the current block of the chunk.
        /// If there are no more records, then make
        /// the next block be current.
        /// If there are no more blocks in the chunk, return false.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/> 
        public bool next()
        {
            while (true)
            {
                if (rp.next())
                    return true;
                if (current == endbnum)
                    return false;
                moveToBlock(current + 1);
            }
        }

        /// <summary>
        /// Closes the current scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close"/> 
        public void close()
        {
            foreach (RecordPage r in pages)
                r.close();
        }

        /// <summary>
        /// Returns the value of the specified field.
        /// </summary>
        ///  <seealso cref="simpledb.query.Scan.getVal"/> 
        public Constant getVal(string fldname)
        {
            if (sch.type(fldname) == Schema.INTEGER)
                return new IntConstant(rp.getInt(fldname));
            else
                return new StringConstant(rp.getString(fldname));
        }

        /// <summary>
        /// Returns the value of the specified field.
        /// </summary>
        ///  <seealso cref="simpledb.query.Scan.getInt"/> 
        public int getInt(string fldname)
        {
            return rp.getInt(fldname);
        }

        /// <summary>
        /// Returns the value of the specified field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getString"/> 
        public string getString(string fldname)
        {
            return rp.getString(fldname);
        }

         /// <summary>
        /// Returns true if the specified field is in
        /// either of the underlying scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/> 
        public bool hasField(string fldname)
        {
            return sch.hasField(fldname);
        }

        private void moveToBlock(int blknum)
        {
            current = blknum;
            rp = pages[current - startbnum];
            rp.moveToId(-1);
        }
    }
}
