﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.tx;
using simpledb.record;
using simpledb.query;

namespace simpledb.multibuffer
{
    /// <summary>
    /// The Scan class for the muti-buffer version of the
    /// <i>product</i> operator.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class MultiBufferProductScan : Scan
    {
        private Scan lhsscan, rhsscan = null, prodscan;
        private TableInfo ti;
        private Transaction tx;
        private int chunksize, nextblknum, filesize;

        /// <summary>
        /// Creates the scan class for the product of the LHS scan and a table.
        /// </summary>
        /// <param name="lhsscan">the LHS scan</param>
        /// <param name="ti">the metadata for the RHS table</param>
        /// <param name="tx">the current transaction</param>
        public MultiBufferProductScan(Scan lhsscan, TableInfo ti, Transaction tx)
        {
            this.lhsscan = lhsscan;
            this.ti = ti;
            this.tx = tx;
            filesize = tx.size(ti.fileName());
            chunksize = BufferNeeds.bestFactor(filesize);
            beforeFirst();
        }

        /// <summary>
        /// Positions the scan before the first record.
        /// That is, the LHS scan is positioned at its first record,
        /// and the RHS scan is positioned before the first record of the first chunk.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.beforeFirst"/> 
        public void beforeFirst()
        {
            nextblknum = 0;
            useNextChunk();
        }

        /// <summary>
        /// Moves to the next record in the current scan.
        /// If there are no more records in the current chunk,
        /// then move to the next LHS record and the beginning of that chunk.
        /// If there are no more LHS records, then move to the next chunk
        /// and begin again.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.next"/>
        public bool next()
        {
            while (!prodscan.next())
                if (!useNextChunk())
                    return false;
            return true;
        }

        /// <summary>
        /// Closes the current scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.close"/> 
        public void close()
        {
            prodscan.close();
        }

        /// <summary>
        /// Returns the value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getVal"/>
        public Constant getVal(string fldname)
        {
            return prodscan.getVal(fldname);
        }

        /// <summary>
        /// Returns the integer value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getInt"/> 
        public int getInt(string fldname)
        {
            return prodscan.getInt(fldname);
        }

        /// <summary>
        /// Returns the string value of the specified field.
        /// The value is obtained from whichever scan
        /// contains the field.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.getString"/> 
        public string getString(string fldname)
        {
            return prodscan.getString(fldname);
        }

        /// <summary>
        /// Returns true if the specified field is in
        /// either of the underlying scans.
        /// </summary>
        /// <seealso cref="simpledb.query.Scan.hasField"/> 
        public bool hasField(string fldname)
        {
            return prodscan.hasField(fldname);
        }

        private bool useNextChunk()
        {
            if (rhsscan != null)
                rhsscan.close();
            if (nextblknum >= filesize)
                return false;
            int end = nextblknum + chunksize - 1;
            if (end >= filesize)
                end = filesize - 1;
            rhsscan = new ChunkScan(ti, nextblknum, end, tx);
            lhsscan.beforeFirst();
            prodscan = new ProductScan(lhsscan, rhsscan);
            nextblknum = end + 1;
            return true;
        }
    }
}
