﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.query;
using simpledb.parse;
using simpledb.planner;
using simpledb.tx;

namespace simpledb.opt
{
    /// <summary>
    /// A query planner that optimizes using a heuristic-based algorithm.
    /// </summary>
    /// <author>Edward Sciore</author>
    class HeuristicQueryPlanner : QueryPlanner
    {
        private List<TablePlanner> tableinfo = new List<TablePlanner>();

        /// <summary>
        /// Creates an optimized left-deep query plan using the following
        /// heuristics.
        /// H1. Choose the smallest table (considering selection predicates)
        /// to be first in the join order.
        /// H2. Add the table to the join order which
        /// results in the smallest output. 
        /// </summary>
        public Plan createPlan(QueryData data, Transaction tx)
        {
            //step 1: create a tablePlanner object for each mentioned table
            foreach (string tblname in data.tables())
            {
                TablePlanner tp = new TablePlanner(tblname, data.pred(), tx);
                tableinfo.Add(tp);
            }

            //step 2: choose the lowest-size plan to begin the join order
            Plan currentplan = getLowestSelectPlan();

            //step 3: Repeatedly add a plan to the join order
            while (tableinfo.Count > 0)
            {
                Plan p = getLowestJoinPlan(currentplan);
                if (p != null)
                    currentplan = p;
                else // no applicable plan
                    currentplan = getLowestProductPlan(currentplan);
            }

            //step4: project on the field names and return 
            return new ProjectPlan(currentplan, data.fields());
        }

        private Plan getLowestSelectPlan()
        {
            TablePlanner besttp = null;
            Plan bestplan = null;
            foreach (TablePlanner tp in tableinfo)
            {
                Plan plan = tp.makeSelectPlan();
                if (bestplan == null || plan.recordsOutput() < bestplan.recordsOutput())
                {
                    besttp = tp;
                    bestplan = plan;
                }
            }
            tableinfo.Remove(besttp);
            return bestplan;
        }

        private Plan getLowestJoinPlan(Plan current)
        {
            TablePlanner besttp = null;
            Plan bestplan = null;
            foreach (TablePlanner tp in tableinfo)
            {
                Plan plan = tp.makeJoinPlan(current);
                if (plan != null && (bestplan == null || plan.recordsOutput() < bestplan.recordsOutput()))
                {
                    besttp = tp;
                    bestplan = plan;
                }
            }
            if (bestplan != null)
                tableinfo.Remove(besttp);
            return bestplan;
        }

        private Plan getLowestProductPlan(Plan current)
        {
            TablePlanner besttp = null;
            TablePlan bestplan = null;
            foreach (TablePlanner tp in tableinfo)
            {
                Plan plan = tp.makeProductPlan(current);
                if (bestplan == null || plan.recordsOutput() < bestplan.recordsOutput())
                {
                    besttp = tp;
                    bestplan = (TablePlan)plan;
                }
            }
            tableinfo.Remove(besttp);
            return bestplan;
        }
    }
}
