﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.tx;
using simpledb.record;
using simpledb.metadata;
using simpledb.query;
using simpledb.server;
using simpledb.index.query;
using simpledb.multibuffer;

namespace simpledb.opt
{
    /// <summary>
    /// This class contains methods for planning a single table.
    /// </summary>
    /// <author>Edward Sciore</author> 
    class TablePlanner
    {
        private TablePlan myplan;
        private Predicate mypred;
        private Schema myschema;
        private Dictionary<string, IndexInfo> indexes;
        private Transaction tx;

        /// <summary>
        /// Creates a new table planner.
        /// The specified predicate applies to the entire query.
        /// The table planner is responsible for determining
        /// which portion of the predicate is useful to the table,
        /// and when indexes are useful.
        /// </summary>
        /// <param name="tblname">the name of the table</param>
        /// <param name="mypred">the query predicate</param>
        /// <param name="tx">the calling transaction</param>
        public TablePlanner(string tblname, Predicate mypred, Transaction tx)
        {
            this.mypred = mypred;
            this.tx = tx;
            myplan = new TablePlan(tblname, tx);
            myschema = myplan.schema();
            indexes = SimpleDB.mdMgr().getIndexInfo(tblname, tx);
        }


        /// <summary>
        /// Constructs a select plan for the table.
        /// The plan will use an indexselect, if possible.
        /// </summary>
        /// <returns>a select plan for the table</returns>
        public Plan makeSelectPlan()
        {
            Plan p = makeIndexSelect();
            if (p == null)
                p = myplan;
            return addSelectPred(p);
        }

        /// <summary>
        /// Constructs a join plan of the specified plan
        /// and the table.  The plan will use an indexjoin, if possible.
        /// (Which means that if an indexselect is also possible,
        /// the indexjoin operator takes precedence.)
        /// The method returns null if no join is possible.
        /// </summary>
        /// <param name="current">the specified plan</param>
        /// <returns>a join plan of the plan and this table</returns>
        public Plan makeJoinPlan(Plan current)
        {
            Schema currsch = current.schema();
            Predicate joinpred = mypred.joinPred(myschema, currsch);
            if (joinpred == null)
                return null;
            Plan p = makeIndexJoin(current, currsch);
            if (p == null)
                p = makeProductJoin(current, currsch);
            return p;
        }

        /// <summary>
        /// Constructs a product plan of the specified plan and
        /// this table.
        /// </summary>
        /// <param name="current">the specified plan</param>
        /// <returns>a product plan of the specified plan and this table</returns>
        public Plan makeProductPlan(Plan current)
        {
            Plan p = addSelectPred(myplan);
            return new MultiBufferProductPlan(current, p, tx);
        }

        private Plan makeIndexSelect()
        {
            foreach (string fldname in indexes.Keys)
            {
                simpledb.query.Constant val = mypred.equatesWithConstant(fldname);
                if (val != null)
                {
                    IndexInfo ii = indexes[fldname];
                    return new IndexSelectPlan(myplan, ii, val, tx);
                }
            }
            return null;
        }

        private Plan makeIndexJoin(Plan current, Schema currsch)
        {
            foreach (string fldname in indexes.Keys)
            {
                string outerfield = mypred.equatsWithField(fldname);
                if (outerfield != null && currsch.hasField(outerfield))
                {
                    IndexInfo ii = indexes[fldname];
                    Plan p = new IndexJoinPlan(current, myplan, ii, outerfield, tx);
                    p = addSelectPred(p);
                    return addJoinPred(p, currsch);
                }
            }
            return null;
        }

        private Plan makeProductJoin(Plan current, Schema currsch)
        {
            Plan p = makeProductPlan(current);
            return addJoinPred(p, currsch);
        }

        private Plan addSelectPred(Plan p)
        {
            Predicate selectpred = mypred.selectPred(myschema);
            if (selectpred != null)
                return new SelectPlan(p, selectpred);
            else
                return p;
        }

        private Plan addJoinPred(Plan p, Schema currsch)
        {
            Predicate joinpred = mypred.joinPred(currsch, myschema);
            if (joinpred != null)
                return new SelectPlan(p, joinpred);
            else
                return p;
        }
    }
}
