﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Text.RegularExpressions;
using System.Collections.Generic;

namespace simpledb.parse
{
    /// <summary>
    /// The lexical analyzer.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class Lexer
    {
        private List<string> keywords = new List<string>();

        private string[] tokens;
        private int position;
        private const string intPattern = @"^[-+]?(0|[1-9]\d*)$";
        private const string stringPattern = @"^'[^']*'$";
        private const string idPattern = @"^[a-zA-Z_]\w*$";
        private const string keywordPattern = @"^[a-zA-Z]*$";
        private const string splitPattern = @"\s{1}|(,){1}|(=){1}|(\(){1}|(\)){1}|('[^']*'){1}";

        /// <summary>
        /// Creates a new lexical analyzer for SQL statement s.
        /// </summary>
        /// <param name="s">the SQL statement</param>
        public Lexer(string s)
        {
            initKeywords();
            tokens = Regex.Split(s, splitPattern);
            position = -1;
            nextToken();
        }

        /// <summary>
        /// Returns true if the current token is
        /// the specified delimiter character.
        /// </summary>
        /// <param name="d">a character denoting the delimiter</param>
        /// <returns>true if the delimiter is the current token</returns>
        public bool matchDelim(char d)
        {
            if (position >= tokens.Length)
                return false;
            return d.ToString() == tokens[position];
        }

        /// <summary>
        /// Returns true if the current token is an integer.
        /// </summary>
        /// <returns>true if the current token is an integer</returns>
        public bool matchIntConstant()
        {
            if (position >= tokens.Length)
                return false;
            return Regex.IsMatch(tokens[position], intPattern);
        }

        /// <summary>
        /// Returns true if the current token is a string.
        /// </summary>
        /// <returns>true if the current token is a string</returns>
        public bool matchstringConstant()
        {
            if (position >= tokens.Length)
                return false;
            return Regex.IsMatch(tokens[position], stringPattern);
        }

        /// <summary>
        /// Returns true if the current token is the specified keyword.
        /// </summary>
        /// <param name="w">w the keyword string</param>
        /// <returns>true if that keyword is the current token</returns>
        public bool matchKeyword(string w)
        {
            if (position >= tokens.Length)
                return false;
            return Regex.IsMatch(tokens[position], keywordPattern) && tokens[position].ToLower() == w;
        }

        /// <summary>
        /// Returns true if the current token is a legal identifier.
        /// </summary>
        /// <returns>true if the current token is an identifier</returns>
        public bool matchId()
        {
            if (position >= tokens.Length)
                return false;
            return Regex.IsMatch(tokens[position], idPattern) && !keywords.Contains(tokens[position].ToLower());
        }

        /// <summary>
        /// Throws an exception if the current token is not the
        /// specified delimiter. 
        /// Otherwise, moves to the next token.
        /// </summary>
        /// <param name="d">d a character denoting the delimiter</param>
        public void eatDelim(char d)
        {
            if (!matchDelim(d))
                throw new BadSyntaxException();
            nextToken();
        }

        /// <summary>
        /// Throws an exception if the current token is not 
        /// an integer. 
        /// Otherwise, returns that integer and moves to the next token.
        /// </summary>
        /// <returns>the integer value of the current token</returns>
        public int eatIntConstant()
        {
            if (!matchIntConstant())
                throw new BadSyntaxException();
            int i = int.Parse(tokens[position]);
            nextToken();
            return i;
        }

        /// <summary>
        /// Throws an exception if the current token is not 
        /// a string. 
        /// Otherwise, returns that string and moves to the next token.
        /// </summary>
        /// <returns>the string value of the current token</returns>
        public string eatStringConstant()
        {
            if (!matchstringConstant())
                throw new BadSyntaxException();
            string s = tokens[position].Substring(1, tokens[position].Length - 2);
            nextToken();
            return s;
        }

        /// <summary>
        /// Throws an exception if the current token is not the
        /// specified keyword. 
        /// Otherwise, moves to the next token.
        /// </summary>
        /// <param name="w">w the keyword string</param>
        public void eatKeyword(string w)
        {
            if (!matchKeyword(w))
                throw new BadSyntaxException();
            nextToken();
        }

        /// <summary>
        /// Throws an exception if the current token is not 
        /// an identifier. 
        /// Otherwise, returns the identifier string 
        /// and moves to the next token.
        /// </summary>
        /// <returns>the string value of the current token</returns>
        public string eatId()
        {
            if (!matchId())
                throw new BadSyntaxException();
            string s = tokens[position].ToLower();
            nextToken();
            return s;
        }

        private void nextToken()
        {
            position++;
            while (position < tokens.Length && tokens[position].Length == 0)
                position++;
        }

        private void initKeywords()
        {
            keywords.Add("select");
            keywords.Add("from");
            keywords.Add("where");
            keywords.Add("and");
            keywords.Add("insert");
            keywords.Add("into");
            keywords.Add("values");
            keywords.Add("delete");
            keywords.Add("update");
            keywords.Add("set");
            keywords.Add("create");
            keywords.Add("table");
            keywords.Add("int");
            keywords.Add("varchar");
            keywords.Add("view");
            keywords.Add("as");
            keywords.Add("index");
            keywords.Add("on");
        }
    }
}
