﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.record;
using simpledb.query;

namespace simpledb.parse
{
    /// <summary>
    /// The SimpleDB parser.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class Parser
    {
        private Lexer lex;

        /// <summary>
        /// Initializes a new instance of the <see cref="Parser"/> class.
        /// </summary>
        /// <param name="s">The SQL sentence.</param>
        public Parser(string s)
        {
            lex = new Lexer(s);
        }

        //some methods for parsing predicates,terms,expressions, constants,and fields.

        private string field()
        {
            return lex.eatId();
        }

        private Constant constant()
        {
            if (lex.matchstringConstant())
                return new StringConstant(lex.eatStringConstant());
            else
                return new IntConstant(lex.eatIntConstant());
        }

        private Expression expression()
        {
            if (lex.matchId())
                return new FieldNameExpression(field());
            else
                return new ConstantExpression(constant());
        }

        private Term term()
        {
            Expression lhs = expression();
            lex.eatDelim('=');
            Expression rhs = expression();
            return new Term(lhs, rhs);
        }

        private Predicate predicate()
        {
            Predicate pred = new Predicate(term());
            if (lex.matchKeyword("and"))
            {
                lex.eatKeyword("and");
                pred.conjoinWith(predicate());
            }
            return pred;
        }

        //Methods for parsing queries

        /// <summary>
        /// Parse SELECT sentence.
        /// </summary>
        /// <returns>a QueryData object</returns>
        public QueryData query()
        {
            lex.eatKeyword("select");
            List<string> fields = selectList();
            lex.eatKeyword("from");
            List<string> tables = tableList();
            Predicate pred = new Predicate();
            if (lex.matchKeyword("where"))
            {
                lex.eatKeyword("where");
                pred = predicate();
            }
            return new QueryData(fields, tables, pred);
        }

        private List<string> selectList()
        {
            List<string> l = new List<string>();
            l.Add(field());
            if (lex.matchDelim(','))
            {
                lex.eatDelim(',');
                l.AddRange(selectList());
            }
            return l;
        }

        private List<string> tableList()
        {
            List<string> l = new List<string>();
            l.Add(lex.eatId());
            if (lex.matchDelim(','))
            {
                lex.eatDelim(',');
                l.AddRange(tableList());
            }
            return l;
        }

        //these are methods for parsing the various update commands

        /// <summary>
        /// Parse INSERT, DELETE, UPDATE, CREATE TABLE, CREATE VIEW, or CREATE INDEX sentence. 
        /// </summary>
        /// <returns>a ?Data object</returns>
        public object updateCmd()
        {
            if (lex.matchKeyword("insert")) 
                return insert();
            else if (lex.matchKeyword("delete"))
                return delete();
            else if (lex.matchKeyword("update"))
                return modify();
            else
                return create();
        }

        private object create()
        {
            lex.eatKeyword("create");
            if (lex.matchKeyword("table"))
                return createTable();
            else if (lex.matchKeyword("view"))
                return createView();
            else
                return createIndex();
        }

        //here are methods for parsing delete command

        private DeleteData delete()
        {
            lex.eatKeyword("delete");
            lex.eatKeyword("from");
            string tblname = lex.eatId();
            Predicate pred = new Predicate();
            if (lex.matchKeyword("where"))
            {
                lex.eatKeyword("where");
                pred = predicate();
            }
            return new DeleteData(tblname, pred);
        }

        // here are methods for parsing insert commands

        private InsertData insert()
        {
            lex.eatKeyword("insert");
            lex.eatKeyword("into");
            string tblname = lex.eatId();
            lex.eatDelim('(');
            List<string> flds = fieldList();
            lex.eatDelim(')');
            lex.eatKeyword("values");
            lex.eatDelim('(');
            List<Constant> vals = constList();
            lex.eatDelim(')');
            return new InsertData(tblname, flds, vals);
        }

        private List<string> fieldList()
        {
            List<string> l = new List<string>();
            l.Add(field());
            if (lex.matchDelim(','))
            {
                lex.eatDelim(',');
                l.AddRange(fieldList());
            }
            return l;
        }

        private List<Constant> constList()
        {
            List<Constant> l = new List<Constant>();
            l.Add(constant());
            if (lex.matchDelim(','))
            {
                lex.eatDelim(',');
                List<Constant> l2 = constList();
                l.AddRange(l2);
            }
            return l;
        }

        //methods for parsing modify commands

        private ModifyData modify()
        {
            lex.eatKeyword("update");
            string tblname = lex.eatId();
            lex.eatKeyword("set");
            string fldname = field();
            lex.eatDelim('=');
            Expression newal = expression();
            Predicate pred = new Predicate();
            if (lex.matchKeyword("where"))
            {
                lex.eatKeyword("where");
                pred = predicate();
            }
            return new ModifyData(tblname, fldname, newal, pred);
        }

        //methods for parsing create table commands

        private CreateTableData createTable()
        {
            lex.eatKeyword("table");
            string tblname = lex.eatId();
            lex.eatDelim('(');
            Schema sch = fieldDefs();
            lex.eatDelim(')');
            return new CreateTableData(tblname, sch);
        }

        private Schema fieldDefs()
        {
            Schema schema = fieldDef();
            if (lex.matchDelim(','))
            {
                lex.eatDelim(',');
                Schema schema2 = fieldDefs();
                schema.addAll(schema2);
            }
            return schema;
        }

        private Schema fieldDef()
        {
            string fldname = field();
            return fieldType(fldname);
        }

        private Schema fieldType(string fldname)
        {
            Schema schema = new Schema();
            if (lex.matchKeyword("int"))
            {
                lex.eatKeyword("int");
                schema.addIntField(fldname);
            }
            else
            {
                lex.eatKeyword("varchar");
                lex.eatDelim('(');
                int strLen = lex.eatIntConstant();
                lex.eatDelim(')');
                schema.addStringField(fldname, strLen);
            }
            return schema;
        }

        // methods for parsing create view commands

        private CreateViewData createView()
        {
            lex.eatKeyword("view");
            string viewname = lex.eatId();
            lex.eatKeyword("as");
            QueryData qd = query();
            return new CreateViewData(viewname, qd);
        }

        //method for parsing create index commands

        private CreateIndexData createIndex()
        {
            lex.eatKeyword("index");
            string idxname = lex.eatId();
            lex.eatKeyword("on");
            string tblname = lex.eatId();
            lex.eatDelim('(');
            string fldname = field();
            lex.eatDelim(')');
            return new CreateIndexData(idxname, tblname, fldname);
        }
    }
}
