// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.tx;

namespace simpledb.record
{
    /// <summary>
    /// Manages a file of records.
    /// There are methods for iterating through the records
    /// and accessing their contents.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class RecordFile
    {
        private TableInfo ti;
        private Transaction tx;
        private string filename;
        private RecordPage rp;
        private int currentblknum;

        /// <summary>
        /// Constructs an object to manage a file of records.
        /// If the file does not exist, it is created.
        /// </summary>
        /// <param name="ti">the table metadata</param>
        /// <param name="tx">the transaction</param>
        public RecordFile(TableInfo ti, Transaction tx) 
        {
            this.ti = ti;
            this.tx = tx;
            filename = ti.fileName();
            if (tx.size(filename) == 0)
                appendBlock();
            moveTo(0);
        }

        /// <summary>
        /// Closes the record file 
        /// </summary>
        public void close()
        {
            rp.close();
        }

        /// <summary>
        /// Positions the current record so that a call to method next
        /// will wind up at the first record. 
        /// </summary>
        public void beforeFirst()
        {
            moveTo(0);
        }

        /// <summary>
        /// Moves to the next record. Returns false if there
        /// is no next record.
        /// </summary>
        /// <returns>false if there is no next record</returns>
        public bool next()
        {
            while (true)
            {
                if (rp.next())
                    return true;
                if (atLastBlock())
                    return false;
                moveTo(currentblknum + 1);
            }
        }

        /// <summary>
        /// Returns the value of the specified field
        /// in the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <returns>the integer value at that field</returns>
        public int getInt(string fldname)
        {
            return rp.getInt(fldname);
        }

        /// <summary>
        /// Returns the value of the specified field
        /// in the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <returns>the string value at that field</returns>
        public string getString(string fldname)
        {
            return rp.getString(fldname);
        }

        /// <summary>
        /// Sets the value of the specified field 
        /// in the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <param name="val">the new value for the field</param>
        public void setInt(string fldname, int val)
        {
            rp.setInt(fldname, val);
        }

        /// <summary>
        /// Sets the value of the specified field 
        /// in the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <param name="val">the new value for the field</param>
        public void setString(string fldname, string val)
        {
            rp.setString(fldname, val);
        }

        /// <summary>
        /// Deletes the current record.
        /// The client must call next() to move to
        /// the next record.
        /// Calls to methods on a deleted record 
        /// have unspecified behavior. 
        /// </summary>
        public void delete()
        {
            rp.delete();
        }

        /// <summary>
        /// Inserts a new, blank record somewhere in the file
        /// beginning at the current record.
        /// If the new record does not fit into an existing block,
        /// then a new block is appended to the file. 
        /// </summary>
        public void insert()
        {
            while (!rp.insert())
            {
                if (atLastBlock())
                    appendBlock();
                moveTo(currentblknum + 1);
            }
        }

        /// <summary>
        /// Positions the current record as indicated by the
        /// specified RID.  
        /// </summary>
        /// <param name="rid">a record identifier</param>
        public void moveToRid(RID rid)
        {
            moveTo(rid.blockNumber());
            rp.moveToId(rid.Id());
        }

        /// <summary>
        /// Returns the RID of the current record.
        /// </summary>
        /// <returns>a record identifier</returns>
        public RID currentRid()
        {
            int id = rp.currentId();
            return new RID(currentblknum, id);
        }

        private void moveTo(int b)
        {
            if (rp != null)
                rp.close();
            currentblknum = b;
            Block blk = new Block(filename, currentblknum);
            rp = new RecordPage(blk, ti, tx);
        }

        private bool atLastBlock()
        {
            return currentblknum == tx.size(filename) - 1;
        }

        private void appendBlock()
        {
            RecordFormatter fmtr = new RecordFormatter(ti);
            tx.append(filename, fmtr);
        }
    }
}
