﻿// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.tx;

namespace simpledb.record
{
    /// <summary>
    /// Manages the placement and access of records in a block.
    /// </summary>
    /// <author>Edward Sciore</author> 
    class RecordPage
    {
        /// <summary>
        /// record status
        /// </summary>
        public const int EMPTY = 0, INUSE = 1;

        private Block blk;
        private TableInfo ti;
        private Transaction tx;
        private int slotsize;
        private int currentslot = -1;

        /// <summary>
        /// Creates the record manager for the specified block.
        /// The current record is set to be prior to the first one.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        /// <param name="ti">the table's metadata</param>
        /// <param name="tx">the transaction performing the operations</param>
        public RecordPage(Block blk, TableInfo ti, Transaction tx)
        {
            this.blk = blk;
            this.ti = ti;
            this.tx = tx;
            slotsize = ti.recordLength() + Page.INT_SIZE;
            tx.pin(blk);
        }

        /// <summary>
        /// Closes the manager, by unpinning the block 
        /// </summary>
        public void close()
        {
            if (blk != null)
            {
                tx.unpin(blk);
                blk = null;
            }
        }

        /// <summary>
        /// Moves to the next record in the block.
        /// </summary>
        /// <returns>false if there is no next record</returns>
        public bool next()
        {
            return searchFor(INUSE);
        }

        /// <summary>
        /// Returns the integer value stored for the 
        /// specified field of the current record.
        /// </summary>
        /// <param name="fldname">The fldname.</param>
        /// <returns></returns>
        public int getInt(string fldname)
        {
            int position = fieldpos(fldname);
            return tx.getInt(blk, position);
        }

        /// <summary>
        /// Returns the integer value stored for the
        /// specified field of the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <returns>the integer stored in that field</returns>
        public string getString(string fldname)
        {
            int position = fieldpos(fldname);
            return tx.getString(blk, position);
        }

        /// <summary>
        /// Stores an integer at the specified field
        /// of the current record. 
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <param name="val">the integer value stored in that field</param>
        public void setInt(string fldname, int val)
        {
            int position = fieldpos(fldname);
            tx.setInt(blk, position, val);
        }
        /// <summary>
        /// Stores a string at the specified field
        /// of the current record.
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <param name="val">the string value stored in that field</param>
        public void setString(string fldname, string val)
        {
            int position = fieldpos(fldname);
            tx.setString(blk, position, val);
        }

        /// <summary>
        /// Deletes the current record.
        /// Deletion is performed by just marking the record
        /// as "deleted"; the current record does not change. 
        /// To get to the next record, call next().
        /// </summary>
        public void delete()
        {
            int position = currentpos();
            tx.setInt(blk, position, EMPTY);
        }

        /// <summary>
        /// Inserts a new, blank record somewhere in the page.
        /// Return false if there were no available slots.
        /// </summary>
        /// <returns>false if the insertion was not possible</returns>
        public bool insert()
        {
            currentslot = -1;
            bool found = searchFor(EMPTY);
            if (found)
            {
                int position = currentpos();
                tx.setInt(blk, position, INUSE);
            }
            return found;
        }

        /// <summary>
        /// Sets the current record to be the record having the
        /// specified ID.  
        /// </summary>
        /// <param name="id">the ID of the record within the page</param>
        public void moveToId(int id)
        {
            currentslot = id;
        }

        /// <summary>
        /// Returns the ID of the current record.
        /// </summary>
        /// <returns>the ID of the current record</returns>
        public int currentId()
        {
            return currentslot;
        }

        private int currentpos()
        {
            return currentslot * slotsize;
        }

         private int fieldpos(string fldname)
        {
            int offset = Page.INT_SIZE + ti.offset(fldname);
            return currentpos() + offset;
        }

        private bool isValidSlot()
        {
            return currentpos() + slotsize <= Page.BLOCK_SIZE;
        }

        private bool searchFor(int flag)
        {
            currentslot++;
            while (isValidSlot())
            {
                int position = currentpos();
                if (tx.getInt(blk, position) == flag)
                    return true;
                currentslot++;
            }
            return false;
        }
    }
}
