// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.file;

namespace simpledb.record
{
    /// <summary>
    /// The metadata about a table and its records.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class TableInfo
    {
        private Schema sch;
        private Dictionary<string, int> offsets;
        private int recordlen;
        private string tblname;

        /// <summary>
        /// Creates a TableInfo object, given a table name
        /// and schema. The constructor calculates the
        /// physical offset of each field.
        /// This constructor is used when a table is created.
        /// </summary>
        /// <param name="tblname">the name of the table</param>
        /// <param name="schema">the schema of the table's records</param>
        public TableInfo(string tblname, Schema schema)
        {
            this.sch = schema;
            this.tblname = tblname;
            offsets = new Dictionary<string, int>();
            int pos = 0;
            foreach (string fldname in schema.fields())
            {
                offsets.Add(fldname, pos);
                pos += lengthInBytes(fldname);
            }
            recordlen = pos;
        }

        /// <summary>
        /// Creates a TableInfo object from the 
        /// specified metadata.
        /// This constructor is used when the metadata
        /// is retrieved from the catalog. 
        /// </summary>
        /// <param name="tblname">the name of the table</param>
        /// <param name="schema">the schema of the table's records</param>
        /// <param name="offsets">the already-calculated offsets of the fields within a record</param>
        /// <param name="recordlen">the already-calculated length of each record</param>
        public TableInfo(string tblname, Schema schema, Dictionary<string, int> offsets, int recordlen)
        {
            this.tblname = tblname;
            this.sch = schema;
            this.offsets = offsets;
            this.recordlen = recordlen;
        }

        /// <summary>
        /// Returns the filename assigned to this table.
        /// Currently, the filename is the table name
        /// followed by ".tbl".
        /// </summary>
        /// <returns>the name of the file assigned to the table</returns>
        public string fileName()
        {
            return tblname + ".tbl";
        }

        /// <summary>
        /// Returns the schema of the table's records
        /// </summary>
        /// <returns>the table's record schema</returns>
        public Schema schema()
        {
            return sch;
        }

        /// <summary>
        /// Returns the offset of a specified field within a record
        /// </summary>
        /// <param name="fldname">the name of the field</param>
        /// <returns>the offset of that field within a record</returns>
        public int offset(string fldname)
        {
            return offsets[fldname];
        }

        /// <summary>
        /// Returns the length of a record, in bytes.
        /// </summary>
        /// <returns>the length in bytes of a record</returns>
        public int recordLength()
        {
            return recordlen;
        }

        private int lengthInBytes(string fldname)
        {
            int fldtype = sch.type(fldname);
            if (fldtype == Schema.INTEGER)
                return Page.INT_SIZE;
            else
                return Page.STR_SIZE(sch.length(fldname));
        }
    }
}
