// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.file;
using simpledb.buffer;
using simpledb.server;

namespace simpledb.tx
{
    /// <summary>
    /// Manages the transaction's currently-pinned buffers.
    /// </summary>
    /// <author>Edward Sciore</author> 
    class BufferList
    {
        private Dictionary<Block, Buffer> buffers = new Dictionary<Block, Buffer>();
        private List<Block> pins = new List<Block>();
        private BufferMgr bufferMgr = SimpleDB.bufferMgr();

        /// <summary>
        /// Returns the buffer pinned to the specified block.
        /// The method returns null if the transaction has not
        /// pinned the block.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        /// <returns>the buffer pinned to that block</returns>
        public Buffer getBuffer(Block blk)
        {
            return buffers.ContainsKey(blk) ? buffers[blk] : null;
        }

        /// <summary>
        /// Pins the block and keeps track of the buffer internally.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void pin(Block blk)
        {
            Buffer buff = bufferMgr.pin(blk);
            if (!buffers.ContainsKey(blk))
                buffers.Add(blk, buff);
            pins.Add(blk);
        }

        /// <summary>
        /// Appends a new block to the specified file
        /// and pins it.
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <param name="fmtr">the formatter used to initialize the new page</param>
        /// <returns>a reference to the newly-created block</returns>
        public Block pinNew(string filename, PageFormatter fmtr)
        {
            Buffer buff = bufferMgr.pinNew(filename, fmtr);
            Block blk = buff.block();
            buffers.Add(blk, buff);
            pins.Add(blk);
            return blk;
        }

        /// <summary>
        /// Unpins the specified block.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void unpin(Block blk)
        {
            Buffer buff = buffers[blk];
            bufferMgr.unpin(buff);
            pins.Remove(blk);
            if (!pins.Contains(blk))
                buffers.Remove(blk);
        }

        /// <summary>
        /// Unpins any buffers still pinned by this transaction. 
        /// </summary>
        public void unpinAll()
        {
            foreach (Block blk in pins)
            {
                Buffer buff = buffers[blk];
                bufferMgr.unpin(buff);
            }
            buffers.Clear();
            pins.Clear();
        }
    }
}
