// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections;
using simpledb.file;
using simpledb.buffer;
using simpledb.tx.concurrency;
using simpledb.tx.recovery;
using simpledb.server;

namespace simpledb.tx
{
    /// <summary>
    /// Provides transaction management for clients,
    /// ensuring that all transactions are serializable, recoverable,
    /// and in general satisfy the ACID properties.
    /// </summary>
    /// <author>Edward Sciore</author>
    public class Transaction
    {
        private static int nextTxNum = 0;
        private const int END_OF_FILE = -1;
        private RecoveryMgr recoveryMgr;
        private ConcurrencyMgr concurMgr;
        private int txnum;
        private BufferList myBuffers = new BufferList();
        private static object threadLock = new object();

        /// <summary>
        /// Creates a new transaction and its associated 
        /// recovery and concurrency managers.
        /// </summary>
        public Transaction()
        {
            txnum = nextTxNumber();
            recoveryMgr = new RecoveryMgr(txnum);
            concurMgr = new ConcurrencyMgr();
        }

        /// <summary>
        /// Commits the current transaction.
        /// Flushes all modified buffers (and their log records),
        /// writes and flushes a commit record to the log,
        /// releases all locks, and unpins any pinned buffers. 
        /// </summary>
        public void commit()
        {
            recoveryMgr.commit();
            concurMgr.release();
            myBuffers.unpinAll();
            //System.Console.WriteLine("transaction " + txnum + " committed");
        }

        /// <summary>
        /// Rolls back the current transaction.
        /// Undoes any modified values,
        /// flushes those buffers,
        /// writes and flushes a rollback record to the log,
        /// releases all locks, and unpins any pinned buffers.
        /// </summary>
        public void rollback()
        {
            recoveryMgr.rollback();
            concurMgr.release();
            myBuffers.unpinAll();
            //System.Console.WriteLine("transaction" + txnum + "rolled back");
        }

        /// <summary>
        /// Flushes all modified buffers.
        /// Then goes through the log, rolling back all
        /// uncommitted transactions.  Finally, 
        /// writes a quiescent checkpoint record to the log.
        /// This method is called only during system startup,
        /// before user transactions begin. 
        /// </summary>
        public void recover()
        {
            SimpleDB.bufferMgr().flushAll(txnum);
            recoveryMgr.recover();
        }

        /// <summary>
        /// Pins the specified block.
        /// The transaction manages the buffer for the client.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void pin(Block blk)
        {
            myBuffers.pin(blk);
        }

        /// <summary>
        /// Unpins the specified block.
        /// The transaction looks up the buffer pinned to this block,
        /// and unpins it. 
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void unpin(Block blk)
        {
            myBuffers.unpin(blk);
        }

        /// <summary>
        /// Returns the integer value stored at the
        /// specified offset of the specified block.
        /// The method first obtains an SLock on the block,
        /// then it calls the buffer to retrieve the value.
        /// </summary>
        /// <param name="blk">a reference to a disk block</param>
        /// <param name="offset">the byte offset within the block</param>
        /// <returns>the integer stored at that offset</returns>
        public int getInt(Block blk, int offset)
        {
            concurMgr.sLock(blk);
            Buffer buff = myBuffers.getBuffer(blk);
            return buff.getInt(offset);
        }

        /// <summary>
        /// Returns the string value stored at the
        /// specified offset of the specified block.
        /// The method first obtains an SLock on the block,
        /// then it calls the buffer to retrieve the value.
        /// </summary>
        /// <param name="blk">a reference to a disk block</param>
        /// <param name="offset">the byte offset within the block</param>
        /// <returns>the string stored at that offset</returns>
        public string getString(Block blk, int offset)
        {
            concurMgr.sLock(blk);
            Buffer buff = myBuffers.getBuffer(blk);
            return buff.getString(offset);
        }

        /// <summary>
        /// Stores an integer at the specified offset 
        /// of the specified block.
        /// The method first obtains an XLock on the block.
        /// It then reads the current value at that offset,
        /// puts it into an update log record, and 
        /// writes that record to the log.
        /// Finally, it calls the buffer to store the value,
        /// passing in the LSN of the log record and the transaction's id
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        /// <param name="offset">a byte offset within that block</param>
        /// <param name="val">the value to be stored</param>
        public void setInt(Block blk, int offset, int val)
        {
            concurMgr.xLock(blk);
            Buffer buff = myBuffers.getBuffer(blk);
            int lsn = recoveryMgr.setInt(buff, offset, val);
            buff.setInt(offset, val, txnum, lsn);
        }

        /// <summary>
        /// Stores a string at the specified offset 
        /// of the specified block.
        /// The method first obtains an XLock on the block.
        /// It then reads the current value at that offset,
        /// puts it into an update log record, and 
        /// writes that record to the log.
        /// Finally, it calls the buffer to store the value,
        /// passing in the LSN of the log record and the transaction's id.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        /// <param name="offset">a byte offset within that block</param>
        /// <param name="val">the value to be stored</param>
        public void setString(Block blk, int offset, string val)
        {
            concurMgr.xLock(blk);
            Buffer buff = myBuffers.getBuffer(blk);
            int lsn = recoveryMgr.setString(buff, offset, val);
            buff.setString(offset, val, txnum, lsn);
        }

        /// <summary>
        /// Returns the number of blocks in the specified file.
        /// This method first obtains an SLock on the 
        /// "end of the file", before asking the file manager
        /// to return the file size.
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <returns>the number of blocks in the file</returns>
        public int size(string filename)
        {
            Block dummyblk = new Block(filename, END_OF_FILE);
            concurMgr.sLock(dummyblk);
            return SimpleDB.fileMgr().size(filename);
        }

        /// <summary>
        /// Appends a new block to the end of the specified file
        /// and returns a reference to it.
        /// This method first obtains an XLock on the
        /// "end of the file", before performing the append.
        /// </summary>
        /// <param name="filename">the name of the file</param>
        /// <param name="fmtr">the formatter used to initialize the new page</param>
        /// <returns>a reference to the newly-created disk block</returns>
        public Block append(string filename, PageFormatter fmtr)
        {
            Block dummyblk = new Block(filename, END_OF_FILE);
            concurMgr.xLock(dummyblk);
            Block blk = myBuffers.pinNew(filename, fmtr);
            unpin(blk);
            return blk;
        }

        private static int nextTxNumber()
        {
            lock (threadLock)
            {
                nextTxNum++;
                //System.Console.WriteLine("new transaction: " + nextTxNum);
                return nextTxNum;
            }

        }
    }
}
