// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System.Collections.Generic;
using simpledb.file;

namespace simpledb.tx.concurrency
{
    /// <summary>
    /// The concurrency manager for the transaction.
    /// Each transaction has its own concurrency manager. 
    /// The concurrency manager keeps track of which locks the 
    /// transaction currently has, and interacts with the
    /// global lock table as needed.
    /// </summary> 
    /// <author>Edward Sciore</author> 
    class ConcurrencyMgr
    {
        /// <summary>
        /// The global lock table.  This variable is static because all transactions
        /// share the same table. 
        /// </summary>
        private static LockTable locktbl = new LockTable();
        private Dictionary<Block, string> locks = new Dictionary<Block, string>();

        /// <summary>
        /// Obtains an SLock on the block, if necessary.
        /// The method will ask the lock table for an SLock
        /// if the transaction currently has no locks on that block.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void sLock(Block blk)
        {
            if (!locks.ContainsKey(blk))
            {
                locktbl.sLock(blk);
                locks.Add(blk, "S");
            }
        }

        /// <summary>
        /// Obtains an XLock on the block, if necessary.
        /// If the transaction does not have an XLock on that block,
        /// then the method first gets an SLock on that block
        /// (if necessary), and then upgrades it to an XLock
        /// </summary>
        /// <param name="blk">a refrence to the disk block</param>
        public void xLock(Block blk)
        {
            if (!hasXLock(blk))
            {
                sLock(blk);
                locktbl.xLock(blk);
                locks[blk] = "X";
            }
        }

        /// <summary>
        /// Releases all locks by asking the lock table to
        /// unlock each one 
        /// </summary>
        public void release()
        {
            foreach (Block blk in locks.Keys)
                locktbl.unlock(blk);
            locks.Clear();
        }

        private bool hasXLock(Block blk)
        {
            return locks.ContainsKey(blk) ? locks[blk].Equals("X") : false;
        }
    }
}
