// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using System;
using System.Collections.Generic;
using System.Threading;
using simpledb.file;

namespace simpledb.tx.concurrency
{
    /// <summary>
    /// The lock table, which provides methods to lock and unlock blocks.
    /// If a transaction requests a lock that causes a conflict with an
    /// existing lock, then that transaction is placed on a wait list.
    /// There is only one wait list for all blocks.
    /// When the last lock on a block is unlocked, then all transactions
    /// are removed from the wait list and rescheduled.
    /// If one of those transactions discovers that the lock it is waiting for
    /// is still locked, it will place itself back on the wait list.
    /// </summary>
    /// <author>Edward Sciore</author>   
    class LockTable
    {
        private const int MAX_TIME = 10000; // 10 seconds
        private object threadLock = new object();
        private Dictionary<Block, int> locks = new Dictionary<Block, int>();

        /// <summary>
        /// Grants an SLock on the specified block.
        /// If an XLock exists when the method is called,
        /// then the calling thread will be placed on a wait list
        /// until the lock is released.
        /// If the thread remains on the wait list for a certain 
        /// amount of time (currently 10 seconds),
        /// then an exception is thrown.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void sLock(Block blk)
        {
            lock (threadLock)
            {
                try
                {
                    long timestamp = DateTime.Now.Ticks;
                    while (hasXlock(blk) && !waitingTooLong(timestamp))
                        Monitor.Wait(threadLock, MAX_TIME);
                    if (hasXlock(blk))
                        throw new LockAbortException();
                    int val = getLockVal(blk); // will not be negative
                    locks[blk] = val + 1;
                }
                catch (ThreadInterruptedException)
                {
                    throw new LockAbortException();
                }
            }
        }

        /// <summary>
        /// Grants an XLock on the specified block.
        /// If a lock of any type exists when the method is called,
        /// then the calling thread will be placed on a wait list
        /// until the locks are released.
        /// If the thread remains on the wait list for a certain 
        /// amount of time (currently 10 seconds),
        /// then an exception is thrown.
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void xLock(Block blk)
        {
            lock (threadLock)
            {
                try
                {
                    long timestamp = DateTime.Now.Ticks;
                    while (hasOtherSLocks(blk) && !waitingTooLong(timestamp))
                        Monitor.Wait(threadLock, MAX_TIME);
                    if (hasOtherSLocks(blk))
                        throw new LockAbortException();
                    locks[blk] = -1;
                }
                catch (ThreadInterruptedException)
                {
                    throw new LockAbortException();
                }
            }
        }

        /// <summary>
        /// Releases a lock on the specified block.
        /// If this lock is the last lock on that block,
        /// then the waiting transactions are notified
        /// </summary>
        /// <param name="blk">a reference to the disk block</param>
        public void unlock(Block blk)
        {
            lock (threadLock)
            {
                int val = getLockVal(blk);
                if (val > 1)
                    locks[blk] = val - 1;
                else
                {
                    locks.Remove(blk);
                    Monitor.PulseAll(threadLock);
                }
            }
        }

        private bool hasXlock(Block blk)
        {
            return getLockVal(blk) < 0;
        }

        private bool hasOtherSLocks(Block blk)
        {
            return getLockVal(blk) > 1;
        }

        private bool waitingTooLong(long starttime)
        {
            return (DateTime.Now.Ticks - starttime) / 10000 > MAX_TIME;
        }

        private int getLockVal(Block blk)
        {
            if (locks.ContainsKey(blk))
                return locks[blk];
            locks.Add(blk, 0);
            return 0;
        }
    }
}
