// migration from java to .NET by SimpleDB# group, 
// http://cs.tju.edu.cn/faculty/zhangkl/projects/simpledbsharp/index.htm

using simpledb.file;
using simpledb.log;
using simpledb.buffer;
using simpledb.server;

namespace simpledb.tx.recovery
{
    /// <summary>
    /// The SetStringRecord log record.
    /// </summary>
    /// <author>Edward Sciore</author> 
    class SetStringRecord : LogRecord
    {
        private int txnum, offset;
        private string val;
        private Block blk;

        /// <summary>
        /// Creates a new setString log record.
        /// </summary>
        /// <param name="txnum">the ID of the specified transaction</param>
        /// <param name="blk">the block containing the value</param>
        /// <param name="offset">the offset of the value in the block</param>
        /// <param name="val">the new value</param>
        public SetStringRecord(int txnum, Block blk, int offset, string val)
        {
            this.txnum = txnum;
            this.blk = blk;
            this.offset = offset;
            this.val = val;
        }

        /// <summary>
        /// Creates a log record by reading five other values from the log.
        /// </summary>
        /// <param name="rec">the basic log record</param>
        public SetStringRecord(BasicLogRecord rec)
        {
            txnum = rec.nextInt();
            string filename = rec.nextString();
            int blknum = rec.nextInt();
            blk = new Block(filename, blknum);
            offset = rec.nextInt();
            val = rec.nextString();
        }

        /// <summary>
        /// Writes a setString record to the log.
        /// This log record contains the setString operator,
        /// followed by the transaction id, the filename, number,
        /// and offset of the modified block, and the previous
        /// string value at that offset.
        /// </summary>
        /// <returns>the LSN of the last log value</returns>
        public override int writeToLog()
        {
            object[] rec = new object[] { SETSTRING, txnum, blk.fileName(), blk.number(), offset, val };
            return logMgr.append(rec);
        }

        /// <summary>
        /// returns the log record's type
        /// </summary>
        /// <returns>the log record's type</returns>
        public override int op()
        {
            return SETSTRING;
        }

        /// <summary>
        /// Returns the transaction id stored with
        /// the log record.
        /// </summary>
        /// <returns>the log record's transaction id</returns>
        public override int txNumber()
        {
            return txnum;
        }

        /// <summary>
        /// Returns a <see cref="System.String"/> that represents this instance.
        /// </summary>
        /// <returns>
        /// A <see cref="System.String"/> that represents this instance.
        /// </returns>
        public override string ToString()
        {
            return "<setString " + txnum + " " + blk + " " + offset + " " + val + ">";
        }

        /// <summary>
        /// Replaces the specified data value with the value saved in the log record.
        /// The method pins a buffer to the specified block,
        /// calls setString to restore the saved value
        /// (using a dummy LSN), and unpins the buffer.
        /// </summary>
        public override void undo(int txnum)
        {
            BufferMgr buffMgr = SimpleDB.bufferMgr();
            Buffer buff = buffMgr.pin(blk);
            buff.setString(offset, val, txnum, -1);
            buffMgr.unpin(buff);
        }
    }
}
